;;; $Id: combinator-tools.scm,v 1.5 2002/04/09 22:25:24 leavens Exp $

;;; The following algorithm is from Hindley, Lercher and Seldin's book
;;;     "Introduction to Combinatory Logic", Cambridge, 1972)

;;; For the input grammar, <combinator-lambda-1-exp>, see the following file:
(load-quietly "combinator-lambda-1-exp.scm")

;;; For the output grammar, <combinator-term>, see the following file:
(load-quietly "combinator-term.scm")

;;; Examples:
;;; (unparse-combinator-term
;;;   (toCombinators (parse-combinator-lambda-1-exp '(lambda (x) (lambda (y) x)))))
;;;  ==> k
;;; (unparse-combinator-term
;;;   (toCombinators (parse-combinator-lambda-1-exp '(lambda (x) (lambda (y) y)))))
;;;  ==> (k i)

(deftype toCombinators (-> (combinator-lambda-1-exp) combinator-term))
(define toCombinators
  (lambda (trm)
    (cases combinator-lambda-1-exp trm
      (var-exp (id) (var id))
      (app-exp (rator rand)
	       (capp (toCombinators rator) (toCombinators rand)))
      (lambda-exp (formal body)
	(if (not (occurs-free-in? formal body))
	    (capp (K) (toCombinators body))
	    (let ((outer-formal formal))
	      (cases combinator-lambda-1-exp body
		(var-exp (id) (I))   ; because formal == id here
		(lambda-exp (inner-formal inner-body)
		  (toCombinators
		   (lambda-exp outer-formal
			       (toLambda (toCombinators body)))))
		(app-exp (rator rand)
		     (if (and (equal? rand (var-exp outer-formal))
			      (not (occurs-free-in? outer-formal rator)))
			 (toCombinators rator)  ; eta rule
			 (capp
			  (capp (S)
				(toCombinators
				 (lambda-exp outer-formal rator)))
			  (toCombinators (lambda-exp outer-formal rand)))))
		(else (toCombinators body))))))
      (LI () (I))
      (LS () (S))
      (LK () (K))
      (else (error "bad combinator-lambda-1-exp term: " trm)))))

(deftype interpret (-> (combinator-term) combinator-term))
(define interpret
  (lambda (trm)
    ;; ENSURES LIBERALLY: result is the normal form of trm
    (displayln (unparse-combinator-term trm) " ==> ")
    (cond
     ((capp-of-I? trm) (interpret (capp->rand trm)))
     ((capp-of-K? trm) (interpret (capp->rand (capp->rator trm))))
     ((capp-of-S? trm)
      (interpret (capp
		  (capp (capp->rand (capp->rator (capp->rator trm)))
			(capp->rand trm))
		  (capp (capp->rand (capp->rator trm))
			(capp->rand trm)))))
     (else trm))))

(deftype capp-of-I? (-> (combinator-term) boolean))
(define capp-of-I?
  (lambda (trm)
    (and (capp? trm)
	 (I? (capp->rator trm)))))

(deftype capp-of-K? (-> (combinator-term) boolean))
(define capp-of-K?
  (lambda (trm)
    (and (capp? trm)
	 (capp? (capp->rator trm))
	 (K? (capp->rator (capp->rator trm))))))

(deftype capp-of-S? (-> (combinator-term) boolean))
(define capp-of-S?
  (lambda (trm)
    (and (capp? trm)
	 (capp? (capp->rator trm))
	 (capp? (capp->rator (capp->rator trm)))
	 (S? (capp->rator (capp->rator (capp->rator trm)))))))


(deftype lambda-calculus-rep (-> () poof))
(define lambda-calculus-rep
  (lambda ()
    ;; EFFECT: read, interpret, and print lambda calculus expressions.
    ;;         You can enter combinator terms directly by using LS, LK, and LI.
    ;;         You can quit by typing in something that doesn't parse, like
    ;;            (quit)
    (display "lambda-calculus> ")
    (force-output)   ; SCM specific
    (pretty-print
     (unparse-combinator-term
      (interpret
       (toCombinators
	(parse-combinator-lambda-1-exp
	 (read))))))
    (newline)
    (lambda-calculus-rep)))

;;; The toLambda function is used internally above;
;;; it does not attempt to recover the initial forma of a lambda term,
;;; which would (in general) be impossible anyway.

(deftype toLambda (-> (combinator-term) combinator-lambda-1-exp))
(define toLambda
  (lambda (ctrm)
    (cases combinator-term ctrm
      (I () (LI))
      (K () (LK))
      (S () (LS))
      (var (symbol) (var-exp symbol))
      (capp (rator rand) (app-exp (toLambda rator) (toLambda rand)))
      )))

(deftype occurs-free-in? (-> (symbol combinator-lambda-1-exp) boolean))
(define occurs-free-in?
  (lambda (sym trm)
    ;; ENSURES: result is true just when sym occurs free in trm
    (cases combinator-lambda-1-exp trm
     (var-exp (id) (eq? sym id))
     (lambda-exp (formal body)
       (and (not (eq? sym formal)) (occurs-free-in? sym body)))
     (app-exp (rator rand)
       (or (occurs-free-in? sym rator)
	   (occurs-free-in? sym rand)))
     (else #f))))
