;;; $Id: lambda-1-quote-exp.scm,v 1.2 2005/02/17 16:29:22 leavens Exp $
;;; An ADT for the untyped 1-argument lambda calculus with quoted symbols
;;;
;;; Grammar:
;;;  <expression> ::=
;;;       <identifier>                               "var-exp (id)"
;;;     | (quote <identifier>)                       "quote-exp (symbol)"
;;;     | ( lambda ( <identifier> ) <expression> )   "lambda-exp (id body)"
;;;     | ( <expression> <expression> )              "app-exp (rator rand)"

(module lambda-1-quote-exp (lib "typedscm.ss" "lib342")

(provide lambda-1-quote-exp? var-exp? quote-exp? lambda-exp? app-exp?
	 var-exp->id quote-exp->symbol lambda-exp->id lambda-exp->body
	 app-exp->rator app-exp->rand
	 var-exp quote-exp lambda-exp app-exp
	 parse-lambda-1-quote-exp)

;; type predicate
(deftype lambda-1-quote-exp? (type-predicate-for lambda-1-quote-exp))

;; case testers (discriminators)
(deftype var-exp? (-> (lambda-1-quote-exp) boolean))
(deftype quote-exp? (-> (lambda-1-quote-exp) boolean))
(deftype lambda-exp? (-> (lambda-1-quote-exp) boolean))
(deftype app-exp? (-> (lambda-1-quote-exp) boolean))

;; observers 
(deftype var-exp->id (-> (lambda-1-quote-exp) symbol))
(deftype quote-exp->symbol (-> (lambda-1-quote-exp) symbol))
(deftype lambda-exp->id (-> (lambda-1-quote-exp) symbol))
(deftype lambda-exp->body (-> (lambda-1-quote-exp) lambda-1-quote-exp))
(deftype app-exp->rator (-> (lambda-1-quote-exp) lambda-1-quote-exp))
(deftype app-exp->rand (-> (lambda-1-quote-exp) lambda-1-quote-exp))

;; constructors
(deftype var-exp (-> (symbol) lambda-1-quote-exp))
(deftype quote-exp (-> (symbol) lambda-1-quote-exp))
(deftype lambda-exp (-> (symbol lambda-1-quote-exp) lambda-1-quote-exp))
(deftype app-exp (-> (lambda-1-quote-exp lambda-1-quote-exp)
		     lambda-1-quote-exp))

;; parsing/bless input correctness
(deftype parse-lambda-1-quote-exp (-> (datum) lambda-1-quote-exp))

(defrep (lambda-1-quote-exp datum))

(define lambda-1-quote-exp?
  (lambda (exp)
    ;; ENSURES: result is true just when d represents a lambda-1-quote-exp
    (has-type-trusted
      boolean
      (or (var-exp? exp)
	  (quote-exp? exp)
	  (and (lambda-exp? exp) 
	       (lambda-1-quote-exp? (lambda-exp->body exp)))
	  (and (app-exp? exp)
	       (lambda-1-quote-exp? (app-exp->rator exp))
	       (lambda-1-quote-exp? (app-exp->rand exp)))))))

(define var-exp?
  (lambda (exp)
    ;; ENSURES: result is true just when exp represents a var-exp
    (symbol? exp)))

(define quote-exp?
  (lambda (exp)
    ;; ENSURES: result is true just when exp represents a quote-exp
    (and (list? exp)
	 (= (length exp) 2)
	 (eq? (car exp) 'quote)
	 (symbol? (cadr exp)))))

(define lambda-exp?
  (lambda (exp)
    ;; ENSURES: result is true just when exp represents a lambda-exp
    (and (list? exp)
	 (= (length exp) 3)
	 (eq? (car exp) 'lambda)
	 (let ((formals (cadr exp)))
	   (and (list? formals)
		(= 1 (length formals))
		(symbol? (car formals)))))))

(define app-exp?
  (lambda (exp)
    ;; ENSURES: result is true just when exp represents an app-exp
    (and (list? exp)
	 (= (length exp) 2)
	 (not (eq? (car exp) 'quote)))))

(define var-exp->id
  (lambda (vr)
    ;; REQUIRES: vr represents a var-exp
    ;; ENSURES: result is the symbol represented by vr
    (if (var-exp? vr)
	vr
	(error "var-exp->id: not a var-exp:" vr))))

(define quote-exp->symbol
  (lambda (qe)
    ;; REQUIRES: qe represents a quote-exp
    ;; ENSURES: result is the operand in the representation of qe
    (if (quote-exp? qe)
	(cadr qe)
	(error "quote-exp->symbol: not a quote-exp:" qe))))

(define lambda-exp->id
  (lambda (lm)
    ;; REQUIRES: lm represents a lambda-exp
    ;; ENSURES: result is the formal in the representation of lm
    (if (lambda-exp? lm)
	(caadr lm)
	(error "lambda-exp->id: not a lambda-exp:" lm))))

(define lambda-exp->body
  (lambda (lm)
    ;; REQUIRES: lm represents a lambda-exp
    ;; ENSURES: result is the body in the representation of lm
    (if (lambda-exp? lm)
	(caddr lm)
	(error "lambda-exp->body: not a lambda-exp:" lm))))

(define app-exp->rator
  (lambda (ap)
    ;; REQUIRES: ap represents an app-exp
    ;; ENSURES: result is the operator in the representation of ap
    (if (app-exp? ap)
	(car ap)
	(error "app-exp->rator: not an app-exp:" ap))))

(define app-exp->rand
  (lambda (ap)
    ;; REQUIRES: ap represents an app-exp
    ;; ENSURES: result is the operand in the representation of ap
    (if (app-exp? ap)
	(cadr ap)
	(error "app-exp->rand: not an app-exp:" ap))))

(define var-exp
  (lambda (id)
    ;; ENSURES: result is a representation of a var-exp with symbol id
    id))

(define quote-exp
  (lambda (sym)
    ;; ENSURES: result is a representation of a quote-exp with symbol sym
    (list 'quote sym)))

(define lambda-exp
  (lambda (id body)
    ;; ENSURES: result is a representation of a lambda-exp with formal id
    ;;          and body body
    (list 'lambda (list id) body)))

(define app-exp
  (lambda (rator rand)
    ;; ENSURES: result is a representation of an app-exp with operator rator
    ;;          and operand rand
    (list rator rand)))

(define parse-lambda-1-quote-exp
  (lambda (exp)
    ;; ENSURES: if exp has the syntax of a lambda-1-quote-exp, then
    ;;          result is the representation of exp as a
    ;;          lambda-1-quote-exp, otherwise an error message is given.
    (has-type-trusted
      lambda-1-quote-exp
      (cond
       ((symbol? exp) (var-exp exp))
       ((quote-exp? exp) (quote-exp (cadr exp)))
       ((lambda-exp? exp)
	(lambda-exp (caadr exp)
		    (parse-lambda-1-quote-exp (caddr exp))))
       ((app-exp? exp)
	(app-exp (parse-lambda-1-quote-exp (car exp))
		 (parse-lambda-1-quote-exp (cadr exp))))
       (else (error "parse-lambda-1-quote-exp: bad syntax for expression: "
		    exp))))))
 ) ; end module
