;;; $Id: sxml-node.scm,v 1.6 2005/03/08 17:59:06 leavens Exp $
;;; AUTHOR: Gary T. Leavens
;;;
;;; Grammar for SXML representation in Scheme, adapted from
;;; http://okmij.org/ftp/Scheme/SXML.html.  The helpers in this file
;;; deal with the abstract syntax of XML parsed by the SXML tools; of
;;; course, XML's concrete syntax itself doesn't look like this.
;;;
;;; <node> ::=
;;;       ( <name> . ( {<node>}* ))   "named (name nodelist)"
;;;     | <string>                    "value (string)"
;;; <name> ::= 
;;;       @                     "attribute ()"
;;;     | *TOP*                 "top ()"
;;;     | *PI*                  "process-instr ()"
;;;     | *COMMENT*             "comment ()"
;;;     | *ENTITY*              "entity ()"
;;;     | *NAMESPACES*          "namespaces ()"
;;;     | <local-name>          "local-name (symbol)"
;;;     | <expanded-name>       "expanded-name (symbol)"
;;;
;;; A more useful set of helpers for SXML is found in sxml-helpers.scm.

(module sxml-node (lib "typedscm.ss" "lib342")

(provide node? name? named? value?
	attribute? top? process-instr? comment? entity? namespaces?
	administrative? local-name? expanded-name?
        named value attribute top process-instr comment entity namespaces
	local-name expanded-name 
        named->name named->nodelist value->string
        local-name->symbol expanded-name->symbol)

;; we rely on an external parser, SSAX, available from the URL above

;; type predicate
(deftype node? (type-predicate-for node))
(deftype name? (type-predicate-for name))

;; case testers (discriminators)
(deftype named? (-> (node) boolean))
(deftype value? (-> (node) boolean))
(deftype attribute? (-> (name) boolean))
(deftype top? (-> (name) boolean))
(deftype process-instr? (-> (name) boolean))
(deftype comment? (-> (name) boolean))
(deftype entity? (-> (name) boolean))
(deftype namespaces? (-> (name) boolean))
(deftype administrative? (-> (name) boolean))
(deftype local-name? (-> (name) boolean))
(deftype expanded-name? (-> (name) boolean))

;; constructors
(deftype named (-> (name (list-of node)) node))
(deftype value (-> (string) node))
(deftype attribute (-> () name))
(deftype top (-> () name))
(deftype process-instr (-> () name))
(deftype comment (-> () name))
(deftype entity (-> () name))
(deftype namespaces (-> () name))
(deftype local-name (-> (symbol) name))
(deftype expanded-name (-> (symbol) name))

;; observers 
(deftype named->name (-> (node) name))
(deftype named->nodelist (-> (node) (list-of node)))
(deftype value->string (-> (node) string))
(deftype local-name->symbol (-> (name) symbol))
(deftype expanded-name->symbol (-> (name) symbol))

;; Representations, see the grammar above for details
(defrep (node datum) (name symbol))

(define node?
  (lambda (d)
    (has-type-trusted
      boolean
      (or (string? d)
	  (and (pair? d)
	       (name? (car d))
	       ((list-of node?) (cdr d)))))))

(define name? symbol?)

(define named?
  (lambda (nd)
    (has-type-trusted
      boolean
      (and (pair? nd) (symbol? (car nd))))))

(define value? string?)


;; The syntax '@ is not technically allowed in Scheme by R^5RS,
;; so we treat it specially.
(define *sxml-helpers:attribsym* (string->symbol "@"))

(define attribute?
  (lambda (nm)
    (eq? nm *sxml-helpers:attribsym*)))

(define top?
  (lambda (nm)
    (eq? nm '*TOP*)))

(define process-instr?
  (lambda (nm)
    (eq? nm '*PI*)))

(define comment?
  (lambda (nm)
    (eq? nm '*COMMENT*)))

(define entity?
  (lambda (nm)
    (eq? nm '*ENTITY*)))

(define namespaces?
  (lambda (nm)
    (eq? nm '*NAMESPACES*)))

(define administrative?
  (lambda (nm)
    (or (attribute? nm)
	(top? nm)
	(process-instr? nm)
	(comment? nm)
	(entity? nm)
	(namespaces? nm))))

(define local-name?
  (lambda (nm)
    (and (not (administrative? nm))
	 (not (sxml-helpers:string-contains? (symbol->string nm) #\:)))))

(define expanded-name?
  (lambda (nm)
    (sxml-helpers:string-contains? (symbol->string nm) #\:)))

(deftype sxml-helpers:string-contains? (-> (string char) boolean))
(define sxml-helpers:string-contains?
  (lambda (str ch)
    (letrec ((iter?
              (lambda (i)
                (and (>= i 0)
                     (or (eqv? (string-ref str i) ch)
                         (iter? (- i 1)))))))
      (iter? (- (string-length str) 1)))))

(define named cons)

(define value
  (lambda (x) x))

(define local-name
  (lambda (x) x))
(define expanded-name
  (lambda (x) x))

(define attribute (lambda () *sxml-helpers:attribsym*))
(define top (lambda () '*TOP*))
(define process-instr (lambda () '*PI*))
(define comment (lambda () '*COMMENT*))
(define entity (lambda () '*ENTITY*))
(define namespaces (lambda () '*NAMESPACES*))

(deftype sxml-helpers:extract-maker
  (forall (T) (-> ((-> (T) boolean) string)
                  (-> (T) datum))))
(define sxml-helpers:extract-maker
  (lambda (test? name)
    (has-type
      (forall (T) (-> (T) datum))
      (lambda (x)
        (if (test? x)
            x
            (error (string-append
                    "expecting a(n) "
                    name
                    " name, but given:")
                   x))))))

(define named->name
  (lambda (nd)
    (car
     (has-type-trusted
       (pair-of symbol (list-of datum))
       ((sxml-helpers:extract-maker named? "named node") nd)))))

(define named->nodelist
  (lambda (nd)
    (cdr
     (has-type-trusted
       (pair-of symbol (list-of datum))
       ((sxml-helpers:extract-maker named? "named node") nd)))))

(define value->string
  (lambda (nd)
    ((sxml-helpers:extract-maker value? "value node") nd)))

(define local-name->symbol
  (lambda (nm)
    ((sxml-helpers:extract-maker local-name? "local name") nm)))

(define expanded-name->symbol
  (lambda (nm)
    ((sxml-helpers:extract-maker expanded-name? "expanded name") nm)))

) ; end module
