;;; $Id: tc-position.scm,v 1.2 2005/03/12 01:53:10 leavens Exp $
;;; Positions in files for reporting errors.

(module tc-position (lib "typedscm.ss" "lib342")

(provide tc:position tc:position? tc:make-token-position tc:make-tree-position
	tc:position->start-line tc:position->end-line tc:position->filename
	tc:position->error-string tc:position->error-lines-string
	tc:current-scanner-position tc:current-line tc:new-line!
	tc:current-input-filename
	tc:start-input-filename! tc:end-input-filename!)

(deftype tc:make-token-position (-> (number string) tc:position))
(deftype tc:make-tree-position (-> (number number string) tc:position))
(deftype tc:position-for-list (-> ((list-of tc:position)) tc:position))
(deftype tc:position->start-line (-> (tc:position) number))
(deftype tc:position->end-line (-> (tc:position) number))
(deftype tc:position->filename (-> (tc:position) string))
(deftype tc:position->error-string (-> (tc:position) string))
(deftype tc:position->error-lines-string (-> (tc:position) string))

(deftype tc:current-scanner-position (-> () tc:position))
(deftype tc:current-line (-> () number))
(deftype tc:new-line! (-> () void))
(deftype tc:current-input-filename (-> () string))
(deftype tc:start-input-filename! (-> (string) void))
(deftype tc:end-input-filename! (-> () void))
(deftype tc:position->start-line (-> (tc:position) number))
(deftype tc:position->end-line (-> (tc:position) number))
(deftype tc:position->filename (-> (tc:position) string))

(define-datatype tc:position tc:position?
  (tc:token-position (line number?)
		     (filename string?))
  (tc:tree-position (start-line number?)
		     (end-line number?)
		     (filename string?)))


(deftype tc:tc-position-initialized boolean)
(define tc:tc-position-initialized #f)

(define tc:tc-position-initialize!
  (lambda ()
    (set! *tc:last-token-position* (tc:token-position 0 "none"))
    (set! *tc:last-tree-position* (tc:tree-position 0 0 "none"))
    (set! tc:tc-position-initialized #t)
    (set! *tc:current-position-stack*
	  (list (tc:make-token-position 1 "<standard input>")))
    ))
    
(deftype *tc:last-token-position* tc:position)
(define *tc:last-token-position* #f)

(define tc:make-token-position
  (lambda (tline tfilename)
    (if (not tc:tc-position-initialized)
	(tc:tc-position-initialize!))
    (cases tc:position *tc:last-token-position*
      (tc:token-position (line filename)
	 (begin
	   (if (or (not (= line tline)) (not (string=? filename tfilename)))
	       (set! *tc:last-token-position*
		     (tc:token-position tline tfilename)))
	   *tc:last-token-position*))
      (else (error "tc:make-token-position: should not happen")))))

(deftype *tc:last-tree-position* tc:position)
(define *tc:last-tree-position* #f)

(define tc:make-tree-position
  (lambda (t-start-line t-end-line tfilename)
    (if (not tc:tc-position-initialized)
	(tc:tc-position-initialize!))
    (cases tc:position *tc:last-tree-position*
      (tc:tree-position (start-line end-line filename)
	 (begin
	   (if (or (not (= start-line t-start-line))
		   (not (= end-line t-end-line))
		   (not (string=? filename tfilename)))
	       (set! *tc:last-tree-position*
		     (tc:tree-position
		      t-start-line t-end-line tfilename)))
	   *tc:last-tree-position*))
      (else (error "tc:make-tree-position: should not happen")))))

(define tc:position-for-list
  (lambda (lp)
    ;; REQUIRES: lp is not null
    (let ((first-pos (car lp))
	  (last-pos (car (reverse lp))))
      (tc:make-tree-position
       (tc:position->start-line first-pos)
       (tc:position->end-line last-pos)
       (tc:position->filename first-pos)))))

(define tc:position->start-line
  (lambda (tp)
    (cases tc:position tp
      (tc:token-position (line filename) line)
      (tc:tree-position (start-line end-line filename) start-line))))

(define tc:position->end-line
  (lambda (tp)
    (cases tc:position tp
      (tc:token-position (line filename) line)
      (tc:tree-position (start-line end-line filename) end-line))))

(define tc:position->filename
  (lambda (tp)
    (cases tc:position tp
      (tc:token-position (line filename) filename)
      (tc:tree-position (start-line end-line filename) filename))))

(define tc:position->error-string
  (lambda (tp)
    (string-append
     (tc:position->filename tp) ": " (tc:position->error-lines-string tp))))

(define tc:position->error-lines-string
  (lambda (tp)
    (cases tc:position tp
      (tc:token-position (line filename)
	(string-append "line " (number->string line) ": "))
      (tc:tree-position (start-line end-line filename)
	(string-append
	 "line"
	 (if (= start-line end-line)
	     (string-append " " (number->string start-line))
	     (string-append "s " (number->string start-line)
			    " to " (number->string end-line)))
	 ": ")))))


(deftype *tc:current-position-stack* (list-of tc:position))
(define *tc:current-position-stack* '())

(define tc:top-current-position
  (lambda ()
    (if (not tc:tc-position-initialized)
	(tc:tc-position-initialize!))
    (if (null? *tc:current-position-stack*)
	(error "tc:top-current-position: null stack")
	(car *tc:current-position-stack*))))

(define tc:pop-current-position
  (lambda ()
    (if (null? *tc:current-position-stack*)
	(error "tc:pop-current-position: null stack")
	(set! *tc:current-position-stack* (cdr *tc:current-position-stack*)))))

(define tc:current-scanner-position
  (lambda ()
    (tc:top-current-position)))

(define tc:current-line
  (lambda ()
    (tc:position->start-line (tc:top-current-position))))

(define tc:new-line!
  (lambda ()
    (if (not tc:tc-position-initialized)
	(tc:tc-position-initialize!))
    (set-car! *tc:current-position-stack*
	      (tc:token-position (+ 1 (tc:current-line))
				      (tc:current-input-filename)))))

(define tc:current-input-filename
  (lambda ()
    (tc:position->filename (tc:top-current-position))))

(define tc:start-input-filename!
  (lambda (filename)
    (set! *tc:current-position-stack*
	  (cons (tc:make-token-position 1 filename)
		*tc:current-position-stack*))))

(define tc:end-input-filename!
  (lambda ()
    (tc:pop-current-position)))
)
