;;; $Id: tc-scheme-parser.scm,v 1.26 2005/03/30 16:40:32 dorn Exp $

;;; A parser for R5RS Scheme.
;;; It includes additions to support EOPL(2e) (define-datatype, cases),
;;; MzScheme modules (module, provide, require)
;;; and our type checker: the defintion form deftype, defreps in modules,
;;; and the expressions has-type and has-type-trusted.
;;; 
;;; BUGS: this doesn't properly handle define-syntax, let-syntax, and
;;; letrec-syntax.  In particular, it never notices a <macro use>.

(module tc-scheme-parser (lib "typedscm.ss" "lib342")

(provide tc:char-seq-form-parser tc:char-seq-program-parser tc:parse-scm-file
	 tc:character-string-form-parser tc:character-string-program-parser
	 tc:character-string-exp-parser
	 *tc:parse-top-level-loads*
	 tc:reset-parsed-file-cache!
	 tc:begin-parsing-file! tc:done-parsing-file!
	 tc:look-in-parsed-file-cache-for
	 tc:remove-parsing-record-for!
	 ;; names from tc-position.scm
	 tc:position tc:position? tc:make-token-position tc:make-tree-position
	 tc:position->start-line tc:position->end-line tc:position->filename
	 tc:position->error-string tc:position->error-lines-string
         ;; names from tc:char-seq-as-stream
         tc:string->char-seq tc:stream->char-seq tc:char-seq-null?
         ;; names from tc:token-seq-as-stream
         tc:make-input-stream
	 )

;; main procedure for interactive use
(deftype tc:char-seq-form-parser (-> (tc:char-seq) tc:scheme-form))

;; main procedures for non-interactive use
(deftype tc:char-seq-program-parser (-> (tc:char-seq) tc:scheme-program))
(deftype tc:character-string-form-parser (-> (string) tc:scheme-form))
(deftype tc:character-string-program-parser (-> (string) tc:scheme-program))
(deftype tc:character-string-exp-parser (-> (string) tc:scheme-exp))

(require (lib "tc-scheme-scanner.scm" "lib342")
         (lib "tc-char-seq-as-stream.scm" "lib342")
	 (lib "tc-token-seq-as-stream.scm" "lib342")
	 (lib "tc-qq-depth-stack.scm" "lib342")
	 (lib "maybe.scm" "lib342")
	 (lib "tc-util.scm" "lib342")
	 (lib "tc-scheme-abstract-syntax.scm" "lib342"))

;; set the following to #f if you don't want to follow top-level loads
(deftype *tc:parse-top-level-loads* boolean)
(define *tc:parse-top-level-loads* #t)

;;; infrastructure (which must appear before the actual Scheme grammar)

(define-datatype tc:parser-answer-record tc:parser-answer?
  (tc:parser-answer
   (position tc:position?) (tree datum?) (unparsed tc:token-seq?)))

(define tc:parser-answer->position
  (lambda (pa)
    (cases tc:parser-answer-record pa
      (tc:parser-answer (position tree unparsed) position))))

(define tc:parser-answer->tree
  (lambda (pa)
    (cases tc:parser-answer-record pa
      (tc:parser-answer (position tree unparsed) tree))))

(define tc:parser-answer->unparsed
  (lambda (pa)
    (cases tc:parser-answer-record pa
      (tc:parser-answer (position tree unparsed) unparsed))))

(deftype tc:scheme-parser-error (-> (tc:position string ...) poof))
(define tc:scheme-parser-error
  (letrec ((blank-before-each
	    (lambda (msgs)
	      (if (null? msgs)
		  ""
		  (string-append " " (car msgs)
				 (blank-before-each (cdr msgs)))))))
    (lambda (position . messages)
      (displayln (tc:position->error-string position)
		 "Syntax error:"
		 (blank-before-each messages))
      (error "Can't recover from this syntax error, parsing aborted!"))))

(deftype tc:unexpected-error (-> ((list-of symbol)) (-> (symbol) poof)))
(define tc:unexpected-error
    (lambda (expected-classes)
      (lambda (actual-token)
	(cases tc:scheme-token-record actual-token
	  (tc:scheme-token (position class data)
	    (tc:scheme-parser-error
	     position
	     "expecting"
	     (if (= 1 (length expected-classes))
		 (string-append
		  "a" (string #\newline) "       "
		  (tc:token-class->string (car expected-classes)))
		 (string-append
		  "one of: " (string #\newline) "      "
		  (apply string-append
			 (map (lambda (s)
				(string-append " " (tc:token-class->string s)))
			      expected-classes))))
	     (string-append (string #\newline) " not")
	     (if (eq? class 'identifier)
		 (symbol->string data)
		 (string-append "a " (tc:token-class->string class)))))
	  (else (error "tc:unexpected-error: bad actual-token"
		       actual-token))))))

(define-datatype tc:buffer-entry-record tc:buffer-entry?
  (tc:buffer-entry
   (position tc:position?) (data datum?)))

(define tc:buffer-entry->position
  (lambda (be)
    (cases tc:buffer-entry-record be
      (tc:buffer-entry (position data) position))))

(define tc:buffer-entry->data
  (lambda (be)
    (cases tc:buffer-entry-record be
      (tc:buffer-entry (position data) data))))

;;; the following adapted from EOPL (first edition) Figure D.1 : pages 472-474

(define tc:check/drop
  (lambda (class next-action)
    (lambda (buffer token-seq)
      (let* ((token (tc:token-seq-head token-seq))
	     (tok-class (tc:scheme-token->class token)))
        (cond
	 ((eq? tok-class class)
	  (next-action buffer (tc:token-seq-tail token-seq)))
	 ((and (eq? tok-class 'identifier)
	       (eq? (tc:scheme-token->data token) class))
	  (next-action buffer (tc:token-seq-tail token-seq)))
	 (else ((tc:unexpected-error (list class)) token)))))))

(define tc:token->buffer-entry
  (lambda (token)
    (cases tc:scheme-token-record token
	   (tc:scheme-token (position class data)
	      (tc:buffer-entry position data)))))

(define tc:check/shift
  (lambda (class next-action)
    (lambda (buffer token-seq)
      (let* ((token (tc:token-seq-head token-seq))
	     (tok-class (tc:scheme-token->class token)))
        (cond
	 ((eq? (tc:scheme-token->class token) class)
	  (next-action
	   (cons (tc:token->buffer-entry token) buffer)
	   (tc:token-seq-tail token-seq)))
	 ((and (eq? tok-class 'identifier)
	       (eq? (tc:scheme-token->data token) class))
	  (next-action
	   (cons (tc:token->buffer-entry token) buffer)
	   (tc:token-seq-tail token-seq)))
	 (else ((tc:unexpected-error (list class)) token)))))))

(define tc:goto-parser-state
  (lambda (state)
    (lambda (buffer token-seq)
      (let ((next-action (state (tc:token-seq-head token-seq))))
        (next-action buffer token-seq)))))

(define tc:reduce
  (lambda (prod-name)
    (lambda (buffer token-seq)
      (tc:parser-answer-maker
       buffer
       token-seq
       (lambda (position reversed-buffer-data)
	 (apply (tc:get-constructor-from-name prod-name)
		(cons position reversed-buffer-data)))))))

(define tc:parser-answer-maker
  (lambda (buffer token-seq tree-proc)
    (let* ((reversed-buffer (reverse buffer))
	   (next-token (tc:token-seq-head token-seq))
	   (default-position (tc:scheme-token->position next-token))
	   (start-line
	    (if (null? reversed-buffer)
		(tc:position->start-line default-position)
		(tc:position->start-line
		 (tc:buffer-entry->position (car reversed-buffer)))))
	   (end-line
	    (if (null? buffer)
		(tc:position->end-line default-position)
		(tc:position->end-line
		 (tc:buffer-entry->position (car buffer)))))
	   (filename
	    (if (null? buffer)
		(tc:position->filename default-position)
		(tc:position->filename
		 (tc:buffer-entry->position (car buffer)))))
	   (pos (tc:make-tree-position start-line end-line filename)))
      (tc:parser-answer
       pos
       (tree-proc pos (map tc:buffer-entry->data reversed-buffer))
       token-seq))))
    
(define tc:emit-list
  (lambda ()
    (lambda (buffer token-seq)
      (tc:parser-answer-maker
       buffer
       token-seq
       (lambda (position reversed-buffer-data)
	 reversed-buffer-data)))))

(define tc:process-nt
  (lambda (state next-action)
    (lambda (buffer token-seq)
      (let ((next-answer ((tc:goto-parser-state state) '() token-seq)))
        (next-action 
          (cons (tc:buffer-entry
		 (tc:parser-answer->position next-answer)
		 (tc:parser-answer->tree next-answer))
		buffer)
          (tc:parser-answer->unparsed next-answer))))))

(define tc:parse-loop*
  (lambda (starters terminators action)
    (letrec
      ((more (lambda (token)
               (let ((class (tc:scheme-token->class token)))
                 (cond
		  ((memq class terminators) (tc:emit-list))
		  ((eq? class 'end-marker)
		   ((tc:unexpected-error (append starters terminators)) token))
		  (else (action (tc:goto-parser-state more))))))))
      more)))

(define tc:parse-loop+
  (lambda (starters terminators action)
    (letrec
      ((more (lambda (token)
               (let ((class (tc:scheme-token->class token)))
                 (cond
		  ((memq class terminators) (tc:emit-list))
		  ((eq? class 'end-marker)
		   ((tc:unexpected-error (append starters terminators)) token))
		  (else (action (tc:goto-parser-state more))))))))
      (lambda (token)
        (action (tc:goto-parser-state more))))))

;;; special versions of the generic parsing infrastructure that use
;;; *tc:qq-depth-stack* to help parse quasiquotations (see R5RS).

(define tc:process-qq-nt!
  (lambda (state next-action)
    (lambda (buffer token-seq)
      (tc:push-new-qq-depth-counter!)
      (let ((next-answer ((tc:goto-parser-state state) '() token-seq)))
	(tc:pop-qq-depth-stack!)
        (next-action 
          (cons (tc:buffer-entry
		 (tc:parser-answer->position next-answer)
		 (tc:parser-answer->tree next-answer))
		buffer)
          (tc:parser-answer->unparsed next-answer))))))

(define tc:inc-qq-depth-and-process-nt!
  (lambda (state next-action)
    (lambda (buffer token-seq)
      (tc:inc-top-of-qq-depth-stack!)
      (let ((next-answer ((tc:goto-parser-state state) '() token-seq)))
        (next-action 
          (cons (tc:buffer-entry
		 (tc:parser-answer->position next-answer)
		 (tc:parser-answer->tree next-answer))
		buffer)
          (tc:parser-answer->unparsed next-answer))))))

(define tc:dec-qq-depth-and-process-nt!
  (lambda (state-if-qq-depth-zero state-if-qq-depth-non-zero next-action)
    (lambda (buffer token-seq)
      (tc:dec-top-of-qq-depth-stack!)
      (let ((next-answer ((tc:goto-parser-state
			   (if (zero? (tc:top-of-qq-depth-stack))
			       state-if-qq-depth-zero
			       state-if-qq-depth-non-zero))
			  '()
			  token-seq)))
	(next-action
	 (cons (tc:buffer-entry
		(tc:parser-answer->position next-answer)
		(tc:parser-answer->tree next-answer))
	       buffer)
	 (tc:parser-answer->unparsed next-answer))))))

(define tc:dec-qq-depth-and-reduce!
  (lambda (prod-name)
    (lambda (buffer token-seq)
      (tc:dec-top-of-qq-depth-stack!)
      (tc:parser-answer-maker
       buffer
       token-seq
       (lambda (position reversed-buffer-data)
	 (apply (tc:get-constructor-from-name prod-name)
		(cons position reversed-buffer-data)))))))

(define tc:inc-qq-depth-and-reduce!
  (lambda (prod-name-if-qq-depth-zero prod-name-if-qq-depth-non-zero)
    (lambda (buffer token-seq)
      (let ((old-qq-depth (tc:top-of-qq-depth-stack)))
	(tc:inc-top-of-qq-depth-stack!)
	(tc:parser-answer-maker
	 buffer
	 token-seq
	 (lambda (position reversed-buffer-data)
	   (apply (tc:get-constructor-from-name
		   (if (zero? old-qq-depth)
		       prod-name-if-qq-depth-zero
		       prod-name-if-qq-depth-non-zero))
		  (cons position reversed-buffer-data))))))))

(deftype tc:check-for-cond-extra-else
  (-> ((list-of tc:scheme-cond-clause)) void))
(define tc:check-for-cond-extra-else
  (lambda (clauses)
    (if (not (null? clauses))
	(cases tc:scheme-cond-clause (car clauses)
	  (tc:else-cond-clause (position sequence)
	    (tc:scheme-parser-error
	     position
	     "more than one else clause in this cond expression"))
	  (else (tc:check-for-cond-extra-else (cdr clauses)))))))

(deftype tc:check-for-case-extra-else
  (-> ((list-of tc:scheme-case-clause)) void))
(define tc:check-for-case-extra-else
  (lambda (clauses)
    (if (not (null? clauses))
	(cases tc:scheme-case-clause (car clauses)
	  (tc:else-case-clause (position sequence)
	    (tc:scheme-parser-error
	     position
	     "more than one else clause in this case expression"))
	  (else (tc:check-for-case-extra-else (cdr clauses)))))))

(deftype tc:check-for-cases-extra-else
  (-> ((list-of tc:scheme-cases-clause)) void))
(define tc:check-for-cases-extra-else
  (lambda (clauses)
    (if (not (null? clauses))
	(cases tc:scheme-cases-clause (car clauses)
	  (tc:else-cases-clause (position sequence)
	    (tc:scheme-parser-error
	     position
	     "more than one else clause in this cases expression"))
	  (else (tc:check-for-cases-extra-else (cdr clauses)))))))

(deftype tc:check-for-arrow-exp-extra-dots
  (-> ((list-of tc:scheme-exp-exp)) void))
(define tc:check-for-arrow-exp-extra-dots
  (lambda (etes)
    (if (not (null? etes))
	(cases tc:scheme-exp (car etes)
	  (tc:varref (position variable)
	    (if (eq? variable '...)
		(tc:scheme-parser-error
		 position
		 "more than one `...' in the arguments of arrow expression")))
	  (else (tc:check-for-arrow-exp-extra-dots (cdr etes)))))))

(deftype tc:check-for-input-type-extra-dots
  (-> ((list-of tc:input-type-exp)) void))
(define tc:check-for-input-type-extra-dots
  (lambda (etes)
    (if (not (null? etes))
	(cases tc:input-type-exp (car etes)
	  (tc:basic-input-type-exp (position identifier)
	    (if (eq? identifier '...)
		(tc:scheme-parser-error
		 position
		 "more than one `...' in the argument types")))
	  (else (tc:check-for-input-type-extra-dots (cdr etes)))))))

(deftype tc:break-out-grouped-defines-in-forms
  (-> ((list-of tc:scheme-form)) (list-of tc:scheme-form)))
(define tc:break-out-grouped-defines-in-forms
  (lambda (forms)
    (if (null? forms)
	'()
	(cases tc:scheme-form (car forms)
	  (tc:definition (position definition)
	    (append (map (lambda (d)
			   (tc:definition
			    (tc:scheme-definition->position d)
			    d))
			 (tc:flatten-definition definition))
		    (tc:break-out-grouped-defines-in-forms (cdr forms))))
	  (else (cons
		 (car forms)
		 (tc:break-out-grouped-defines-in-forms (cdr forms))))))))

(deftype tc:flatten-definition
  (-> (tc:scheme-definition) (list-of tc:scheme-definition)))
(define tc:flatten-definition
  (lambda (definition)
    ;; ENSURES: Result is a list of definitions containing just those
    ;; in definition, but such that there are no grouped-defines in
    ;; the result.
    (cases tc:scheme-definition definition
      (tc:grouped-define (position definitions)
	(apply append (map tc:flatten-definition definitions)))
      (else (list definition)))))

(deftype tc:extract-module-required-module-names
  (-> (tc:position (list-of tc:scheme-definition)) (list-of tc:module-name)))
(define tc:extract-module-required-module-names
  (letrec ((find-require
	    (lambda (forms)
	      ;; ENSURES: result is a list of module-names in the require
	      (if (null? forms)
		  '()
		  (cases tc:scheme-definition (car forms)
		    (tc:require (position module-names) module-names)
		    (tc:require-for-syntax (position module-names)
		       (find-require (cdr forms)))
		    (tc:deftype (position name type)
	               (find-require (cdr forms)))
		    (else '()))))))
  (lambda (forms)
    ;; ENSURES: result is a list of module-names in the require, if any,
    ;; in the module, or the empty list if there are none.
    (find-require forms))))

(deftype tc:extract-module-required-for-syntax-module-names
  (-> (tc:position (list-of tc:scheme-definition)) (list-of tc:module-name)))
(define tc:extract-module-required-for-syntax-module-names
  (letrec ((find-require
	    (lambda (forms)
	      ;; ENSURES: result is a list of module-names in the require
	      (if (null? forms)
		  '()
		  (cases tc:scheme-definition (car forms)
		    (tc:require-for-syntax (position module-names) module-names)
		    (tc:deftype (position name type)
	               (find-require (cdr forms)))
		    (else '()))))))
  (lambda (forms)
    ;; ENSURES: result is a list of module-names in the require, if any,
    ;; in the module, or the empty list if there are none.
    (find-require forms))))

(deftype tc:extract-module-defrep-pairs
  (-> (tc:position (list-of tc:scheme-definition)) (list-of tc:scheme-abs-conc-pair)))
(define tc:extract-module-defrep-pairs
  (letrec
      ((find-defrep
	(lambda (forms seen-require seen-deftype-after-require seen-require-for-syntax)
	  ;; ENSURES: result is the abs-conc-pairs in the defrep, if any,
	  ;; or the empty list if none.  Also check that there is
	  ;; just one require, and that it comes after any deftypes.
	  (if (null? forms)
	      '()
	      (cases tc:scheme-definition (car forms)
		(tc:defrep (position pairs)
		  (if (not seen-deftype-after-require)
		      pairs
		      (tc:scheme-parser-error
		       position
		       "No deftypes are allowed between"
		       "the preceeding require and this defrep")))
		(tc:require (position module-names)
		  (if (not seen-require)
		      (begin
			(set! seen-require #t)
			(find-defrep (cdr forms) seen-require
				     seen-deftype-after-require seen-require-for-syntax))
		      (tc:scheme-parser-error
		       position
		       "Two require forms in this module."
		       "Only one require form is allowed per module."
		       "You can write several module-names in one require form,"
		       "like (require \"file1.scm\" \"file2.scm\").")))
		(tc:require-for-syntax (position module-names)
		  (if seen-require
		      (tc:scheme-parser-error
		       position "No require-for-syntax allowed after require form.")
		      (if seen-require-for-syntax
			  (tc:scheme-parser-error
			   position "Only one require-for-syntax allowed per module.")
			  (begin
			    (set! seen-require-for-syntax #t)
			    (find-defrep (cdr forms) seen-require
					 seen-deftype-after-require seen-require-for-syntax)))))
		(tc:deftype (position name type)
		  (if (or seen-require seen-require-for-syntax)
		      (set! seen-deftype-after-require #t))
		  (find-defrep (cdr forms)
			       seen-require seen-deftype-after-require seen-require-for-syntax))
		(else '()))))))
    (lambda (forms)
      ;; ENSURES: result is a list of abs-conc-pairs in the defrep, if any,
      ;; in the module, or the empty list if there are none.
      (find-defrep forms #f #f #f))))

(deftype tc:extract-module-definitions
  (-> (tc:position (list-of tc:scheme-definition)) (list-of tc:scheme-definition)))
(define tc:extract-module-definitions
  (letrec
      ((find-defs
	(lambda (forms acc seen-defrep)
	  ;; ENSURES: Accumulate the definitions, ignoring the
	  ;; require, require-for-syntax, defrep, and deftype if any.
	  ;; Also check that there is just one defrep that comes
	  ;; after any deftypes and an optional require
	  (if (null? forms)
	      acc
	      (cases tc:scheme-definition (car forms)
		(tc:deftype (position variable type)
		   (find-defs (cdr forms) acc seen-defrep))
		(tc:require (position pairs)
		   (if (not seen-defrep)		       
		       (find-defs (cdr forms) acc seen-defrep)
		       (tc:scheme-parser-error
			position
			"No require forms are allowed after the defrep.")))
		(tc:defrep (position module-names)
		   (if (not seen-defrep)
		       (find-defs (cdr forms) acc #t)
		       (tc:scheme-parser-error
			position
			"Two defrep forms in this module."
			"Only one defrep form is allowed per module."
			"You can write several module-names in one require form,"
			"like (defrep (abst1 conct1) (abst2 conct2)).")))
		(tc:require-for-syntax (position module-names)
		   (if (not seen-defrep)		       
		       (find-defs (cdr forms) acc seen-defrep)
		       (tc:scheme-parser-error
			position
			"No require-for-syntax forms are allowed after the defrep.")))
;		   (find-defs (cdr forms) acc seen-defrep))
		(else
		 (find-defs
		  (cdr forms)
		  (append acc (tc:flatten-definition (car forms)))
		  seen-defrep)))))))
    (lambda (forms)
      ;; ENSURES: result is a list of the definition forms, in order,
      ;; in forms, that is without the require, deftype, and defrep.
      (find-defs forms '() #f))))

(deftype tc:extract-module-deftypes
  (-> (tc:position (list-of tc:scheme-definition)) (list-of tc:scheme-definition)))
(define tc:extract-module-deftypes
  (letrec
      ((find-deftypes
	(lambda (forms acc)
	  (if (null? forms)
	      acc
	      (cases tc:scheme-definition (car forms)
		   (tc:deftype (position variable type)
			       (find-deftypes
				(cdr forms)
				(append acc (tc:flatten-definition (car forms)))))
		   (else
		    (find-deftypes (cdr forms) acc)))))))
    (lambda (forms)
      ;; ENSURES: result is a list of the deftype forms, in order in forms
      (find-deftypes forms '()))))

(deftype tc:extract-forms-deftypes
  (-> ((list-of tc:scheme-form)) (list-of tc:scheme-definition)))
(define tc:extract-forms-deftypes
  (lambda (forms)
    (tc:extract-module-deftypes
     (map tc:definition->definition
	  (tc:filter tc:definition? forms)))))

(define tc:non-deftype-form?
  (lambda (form)
    (not (and (tc:definition? form)
	      (tc:deftype? (tc:definition->definition form))))))

(define tc:get-constructor-from-name
  (lambda (prod-name)
    (case prod-name
      ;; programs
      ((scheme-program-normal)
       (lambda (pos forms)
	 (let ((forms (tc:break-out-grouped-defines-in-forms (map tc:parse-loads forms))))	 
	   (tc:scheme-program-normal
	    pos
	    (tc:filter tc:require-form? forms)
	    (tc:extract-forms-deftypes forms)
	    (tc:filter (lambda (x)
			 (and (tc:non-deftype-form? x)
			      (not (tc:require-form? x))))
		       forms)))))
      ;; forms (top level commands and definitions)
      ((command) tc:command)
      ;;((load-command) tc:load-command)
      ((definition) tc:definition)
      ((syntax-definition) tc:syntax-definition)
      ((begin-form)
       (lambda (pos forms)
	 (if (tc:all tc:definition? forms)
	     (tc:definition
	      pos
	      (tc:grouped-define
	       pos
	       (map tc:definition->definition forms)))
	     (tc:begin-form
	      pos
	      (tc:break-out-grouped-defines-in-forms forms)))))
      ((require-form) tc:require-form)
      ((module)
       (lambda (pos identifier initial-required-module-name
		    provide-form body-forms)
         ;; (displayln "module's BODY-FORMS = " body-forms)
	 (let ((required-module-names
		(tc:extract-module-required-module-names
		 body-forms))
	       (required-for-syntax-module-names
		(tc:extract-module-required-for-syntax-module-names
		 body-forms))
	       (defrep-pairs
		 (tc:extract-module-defrep-pairs body-forms))
	       (deftypes (tc:extract-module-deftypes body-forms))
	       (defs (tc:extract-module-definitions body-forms)))
	 (tc:module pos identifier initial-required-module-name
		    (tc:provide->exported-ids provide-form)
		    required-for-syntax-module-names
		    required-module-names
		    defrep-pairs
		    deftypes
		    defs))))
      ;; module-body-forms
      ;; Note that tc:provide, tc:require and tc:defrep don't have an
      ;; independent life outside the parser, and are used as
      ;; intemediate forms.  Their information ends up as part of a tc:module.
      ;; require-for-syntax forms are also ignored by the parser.
      ((provide) tc:provide)
      ((require) tc:require) ; not the same as require-form
      ((defrep) tc:defrep)
      ((require-for-syntax) tc:require-for-syntax)
      ;; module-names
      ((module-id) tc:module-id)
      ((unix-relative-path-string) tc:unix-relative-path-string)
      ((lib) tc:lib)
      ((file) tc:file)
      ;; abs-conc-pairs, used in defreps
      ((abs-conc-pair-simple) tc:abs-conc-pair-simple)
      ((abs-conc-pair-polymorphic) tc:abs-conc-pair-polymorphic)
      ;; definitions
      ((simple-define)
       (lambda (pos variable exp)
	 (cases tc:scheme-exp exp
	    (tc:lambda-exp
	     (position formals body)
	     (tc:simple-define
	      pos variable
	      (tc:named-lambda-exp pos variable formals body)))
	    (else (tc:simple-define pos variable exp)))))
      ((procedure-define) tc:procedure-define)
      ;; tc:grouped-define is used during parsing of modules,
      ;; but should not escape the parser
      ((grouped-define) tc:grouped-define)
      ((define-datatype) tc:define-datatype)
      ((deftype) tc:deftype)
      ;; formals
      ((normal-formals) tc:normal-formals)
      ((unrestricted-formals) tc:unrestricted-formals)
      ((some-normal-formals) tc:some-normal-formals)
      ;; bodies
      ((body)
       (lambda (pos forms-list)
	 (tc:apply-to-pos-defs-and-sequence
	  pos forms-list tc:scheme-body)))
      ((begin-def-comm)
       (lambda (pos forms)
	 (if (tc:all tc:definition? forms)
	     (tc:definition
	      pos
	      (tc:grouped-define
	       pos
	       (map tc:definition->definition forms)))
	     (let ((reversed-exps (reverse (map tc:checked-form->exp forms))))
	       (tc:command
		pos
		(tc:begin-exp pos
				   (reverse (cdr reversed-exps))
				   (car reversed-exps)))))))
      ;; variant-decls
      ((variant-decl) tc:variant-decl)
      ;; field-pred-bindings
      ((field-pred-binding) tc:field-pred-binding)
      ;; scheme-syntax-specs
      ((scheme-syntax-spec) tc:scheme-syntax-spec)
      ;; scheme-transformer-specs
      ((scheme-transformer-spec) tc:scheme-transformer-spec)
      ;; syntax-rules
      ((scheme-syntax-rule) tc:scheme-syntax-rule)
      ;; syntax patterns
      ((syntax-pattern-identifier) tc:syntax-pattern-identifier)
      ((syntax-pattern-normal-list)
       (lambda (pos patterns)
	 (if (<= (length patterns) 1)
	     (tc:syntax-pattern-normal-list pos patterns)
	     (let ((reversed-patterns (reverse patterns)))
	       (cases tc:scheme-syntax-pattern (car reversed-patterns)
		 (tc:syntax-pattern-identifier (position identifier)
		    (if (eq? identifier '...)
			(tc:syntax-pattern-ellipsis-list
			 pos
			 (reverse (cddr reversed-patterns))
			 (cadr reversed-patterns))
			(tc:syntax-pattern-normal-list pos patterns)))
		 (else
		  (tc:syntax-pattern-normal-list pos patterns)))))))
      ;;((syntax-pattern-ellipsis-list) tc:syntax-pattern-ellipsis-list)
      ((syntax-pattern-dotted-list)
       (lambda (pos before-dot after-dot)
	 (if (null? before-dot)
	     (tc:scheme-parser-error
	      pos
	      "missing <pattern> before the dot (.)")
	     (tc:syntax-pattern-dotted-list pos before-dot after-dot))))
      ((syntax-pattern-normal-vector)
       (lambda (pos patterns)
	 (if (<= (length patterns) 1)
	     (tc:syntax-pattern-normal-vector pos patterns)
	     (let ((reversed-patterns (reverse patterns)))
	       (cases tc:scheme-syntax-pattern (car reversed-patterns)
		 (tc:syntax-pattern-identifier (position identifier)
		    (if (eq? identifier '...)
			(tc:syntax-pattern-ellipsis-vector
			 pos
			 (reverse (cddr reversed-patterns))
			 (cadr reversed-patterns))
			(tc:syntax-pattern-normal-vector pos patterns)))
		 (else
		  (tc:syntax-pattern-normal-vector pos patterns)))))))
      ;;((syntax-pattern-ellipsis-vector)
      ;;  tc:syntax-pattern-ellipsis-vector)
      ((syntax-pattern-datum) tc:syntax-pattern-datum)
      ;; syntax templates
      ((syntax-template-identifier) tc:syntax-template-identifier)
      ((syntax-template-normal-list) tc:syntax-template-normal-list)
      ((syntax-template-dotted-list)
       (lambda (pos before-dot after-dot)
	 (if (null? before-dot)
	     (tc:scheme-parser-error
	      pos
	      "missing <template element> before the dot (.)")
	     (tc:syntax-template-dotted-list pos before-dot after-dot))))
      ((syntax-template-normal-vector) tc:syntax-template-normal-vector)
      ((syntax-template-datum) tc:syntax-template-datum)
      ;; syntax template elements
      ((scheme-syntax-template-element-normal)
       tc:scheme-syntax-template-element-normal)
      ((scheme-syntax-template-element-ellipsis)
       tc:scheme-syntax-template-element-ellipsis)
      ;; expressions
      ((varref) tc:varref)
      ((quotation) tc:quotation)
      ((self-evaluating) tc:self-evaluating)
      ((empty-list)
       (lambda (pos) (tc:self-evaluating pos '())))
      ((procedure-call)
       (lambda (pos exps)
	 (tc:procedure-call pos (car exps) (cdr exps))))
      ((lambda-exp) tc:lambda-exp)
      ((conditional-exp) tc:conditional-exp)
      ((conditional-stmt) tc:conditional-stmt)
      ((assignment) tc:assignment)
      ((macro-use) tc:macro-use)
      ((let-syntax-exp) tc:let-syntax-exp)
      ((letrec-syntax-exp) tc:letrec-syntax-exp)
      ((cond)
       (lambda (pos clauses)
	 ;; ASSUME: (length clauses) >= 1, because the parser uses a + loop
	 (let ((reversed-clauses (reverse clauses)))
	   (tc:check-for-cond-extra-else (cdr reversed-clauses))
	   (cases tc:scheme-cond-clause (car reversed-clauses)
	     (tc:else-cond-clause (position sequence)
		(tc:cond-exp pos clauses))
	     (else (tc:cond-stmt pos clauses))))))
      ;; ((cond-exp) tc:cond-exp)
      ;; ((cond-stmt) tc:cond-stmt)
      ((case)
       (lambda (pos exp clauses)
	 ;; ASSUME: (length cases) >= 1, because the parser uses a + loop
	 (let ((reversed-clauses (reverse clauses)))
	   (tc:check-for-case-extra-else (cdr reversed-clauses))
	   (cases tc:scheme-case-clause (car reversed-clauses)
	     (tc:else-case-clause (position sequence)
		(tc:case-exp pos exp clauses))
	     (else (tc:case-stmt pos exp clauses))))))
      ;; ((case-exp) tc:case-exp)
      ;; ((case-stmt) tc:case-stmt)
      ((and-exp) tc:and-exp)
      ((or-exp) tc:or-exp)
      ((begin-exp)
       (lambda (pos exps)
	 (let* ((reversed-exps (reverse exps))
		(commands (reverse (cdr reversed-exps)))
		(exp (car reversed-exps)))
	   (tc:begin-exp pos commands exp))))
      ((sequence)
       (lambda (pos exps)
	 (let* ((reversed-exps (reverse exps))
		(commands (reverse (cdr reversed-exps)))
		(exp (car reversed-exps)))
	   (if (null? commands)
	       exp
	       (tc:begin-exp pos commands exp)))))
      ((let-exp) tc:let-exp)
      ((named-let-exp) tc:named-let-exp)
      ((let*-exp) tc:let*-exp)
      ((letrec-exp) tc:letrec-exp)
      ((do-exp) tc:do-exp)
      ((do-stmt) tc:do-stmt)
      ((delay-exp) tc:delay-exp)
      ((quasiquotation) tc:quasiquotation)
      ((cases)
       (lambda (pos type-name exp clauses)
	 ;; ASSUME: (length clauses) >= 1, because the parser uses a + loop
	 (let ((reversed-clauses (reverse clauses)))
	   (tc:check-for-cases-extra-else (cdr reversed-clauses))
	   (cases tc:scheme-cases-clause (car reversed-clauses)
	     (tc:else-cases-clause (position sequence)
		(tc:cases-exp pos type-name exp clauses))
	     (tc:normal-cases-clause (position variant-name fields sequence)
		(tc:cases-stmt pos type-name exp clauses))))))
      ;;((cases-exp) tc:cases-exp)
      ;;((cases-stmt) tc:cases-stmt)
      ((has-type-exp) tc:has-type-exp)
      ((has-type-trusted-exp) tc:has-type-trusted-exp)
      ((test-type-exp) tc:test-type-exp)
      ((arrow-exp)
       (lambda (pos arg-exps result-exp)
	 (let ((reversed-arg-exps (reverse arg-exps))
	       (len (length arg-exps)))
	   (cond
	    ((= len 0)
	     (tc:arrow-exp pos arg-exps result-exp))
	    ((= len 1)
	     (cases tc:scheme-exp (car arg-exps)
	       (tc:varref (position variable)
		 (if (eq? variable '...)
		     (tc:scheme-parser-error
		      position
		      "missing <expression> before `...' in argument expressions")
		     (tc:arrow-exp
		      pos arg-exps result-exp)))
	       (else (tc:arrow-exp
		      pos arg-exps result-exp))))
	    (else
	     (cases tc:scheme-exp (car reversed-arg-exps)
	       (tc:varref (position variable)
		 (if (eq? variable '...)
		     (begin
		       (tc:check-for-arrow-exp-extra-dots
			(cdr reversed-arg-exps))
		       (tc:varargs-arrow-exp
			pos
			(reverse (cddr reversed-arg-exps))
			(cadr reversed-arg-exps)
			result-exp))
		     (tc:arrow-exp pos arg-exps result-exp)))
	       (else (tc:arrow-exp pos arg-exps result-exp))))))))
      ;;((varargs-arrow-exp) tc:varargs-arrow-exp)
      ((forall-exp) tc:forall-exp)
      ;; the following are only used internally, not part of external syntax
      ;;((define-datatype-exp) tc:define-datatype-exp)
      ;;((constructor-for-variant-type-exp)tc:constructor-for-variant-type-exp)
      ;;
      ;; scheme-cond-clause
      ((normal-cond-clause) tc:normal-cond-clause)
      ((test-only-cond-clause) tc:test-only-cond-clause)
      ((test=>cond-clause) tc:test=>cond-clause)
      ((else-cond-clause) tc:else-cond-clause)
      ;; scheme-case-clause
      ((normal-case-clause) tc:normal-case-clause)
      ((else-case-clause) tc:else-case-clause)
      ;; scheme-binding-spec (for let, let*, letrec)
      ((scheme-binding-spec)
       (lambda (pos variable exp)
	 (cases tc:scheme-exp exp
	   (tc:lambda-exp (position formals body)
	      (tc:scheme-binding-spec
	       pos variable
	       (tc:named-lambda-exp pos variable formals body)))
	   (else (tc:scheme-binding-spec pos variable exp)))))
      ;; scheme-iteration-spec (for do)
      ((stepped-iteration-spec) tc:stepped-iteration-spec)
      ((init-only-iteration-spec) tc:init-only-iteration-spec)
      ;; scheme-cases-clause
      ((normal-cases-clause) tc:normal-cases-clause)
      ((else-cases-clause) tc:else-cases-clause)
      ;; scheme-datum (for quotations and quasiquotations)
      ((simple-datum) tc:simple-datum)
      ((normal-list-datum) tc:normal-list-datum)
      ((improper-list-datum) 
       (lambda (pos before-dot after-dot)
	 ;; We massage the before-dot list to make it easier to typecheck:
	 ;; We always desugar (x y z . a) to (x . (y . (z . a)))
	 (letrec ((process-reversed-elements
		   (lambda (reversed-elems cdr-val)
		     (if (not (null? reversed-elems))
			 (process-reversed-elements
			  (cdr reversed-elems)
			  (tc:improper-list-datum pos (list (car reversed-elems)) cdr-val))
			 cdr-val))))
	   (process-reversed-elements (reverse before-dot) after-dot))))
      ((quoted-datum) tc:quoted-datum)
      ((quasiquoted-datum) tc:quasiquoted-datum)
      ((unquoted-datum) tc:unquoted-datum)
      ((unquoted-exp) tc:unquoted-exp)
      ((unquote-splicing-datum) tc:unquote-splicing-datum)
      ((unquote-splicing-exp) tc:unquote-splicing-exp)
      ((vector-datum) tc:vector-datum)
      ;; type expressions added for the type checker
      ((basic-input-type-exp) tc:basic-input-type-exp)
      ((function-input-type-exp)
       (lambda (pos arg-types result-type)
	 (let ((reversed-arg-types (reverse arg-types))
	       (len (length arg-types)))
	   (cond
	    ((= len 0)
	     (tc:function-input-type-exp pos arg-types result-type))
	    ((= len 1)
	     (cases tc:input-type-exp (car arg-types)
	       (tc:basic-input-type-exp (position identifier)
		 (if (eq? identifier '...)
		     (tc:scheme-parser-error
		      position
		      "missing <type exp> before `...' in argument types")
		     (tc:function-input-type-exp
		      pos arg-types result-type)))
	       (else (tc:function-input-type-exp
		      pos arg-types result-type))))
	    (else
	     (cases tc:input-type-exp (car reversed-arg-types)
	       (tc:basic-input-type-exp (position identifier)
		 (if (eq? identifier '...)
		     (begin
		       (tc:check-for-input-type-extra-dots
			(cdr reversed-arg-types))
		       (tc:varargs-function-input-type-exp
			pos
			(reverse (cddr reversed-arg-types))
			(cadr reversed-arg-types)
			result-type))
		     (tc:function-input-type-exp
		      pos arg-types result-type)))
	       (else (tc:function-input-type-exp
		      pos arg-types result-type))))))))
      ;;((varargs-function-input-type-exp)
      ;;  tc:varargs-function-input-type-exp)
      ((intersection-input-type-exp) tc:intersection-input-type-exp)
      ((type-predicate-for-input-type-exp) tc:type-predicate-for-input-type-exp)
      ((variant-record-input-type-exp) tc:variant-record-input-type-exp)
      ((applied-input-type-exp) tc:applied-input-type-exp)
      ((polymorphic-input-type-exp) tc:polymorphic-input-type-exp)
      ;; variant-type-decls
      ((variant-type-decl) tc:variant-type-decl)
      ;; field-type-bindings
      ((field-type-binding) tc:field-type-binding)
      (else (error "tc:get-constructor-from-name: unknown production name:"
		   prod-name)))))

(define tc:checked-form->exp
  (lambda (form)
    (cases tc:scheme-form form
      (tc:command (position exp) exp)
      (else (tc:scheme-parser-error
	     (tc:scheme-form->position form)
	     "expecting an <expression> in this <body>")))))

(define tc:apply-to-pos-defs-and-sequence
  (lambda (pos forms-list proc)
    (letrec ((find-end-of-defs
	      (lambda (defs forms-list)
		(cond
		 ((null? forms-list)
		  (tc:scheme-parser-error
		   pos
		   "missing <expression> in <body> starting here"))
		 ((tc:definition? (car forms-list))
		  (find-end-of-defs
		   (append
		    (reverse
		     (tc:flatten-definition
		      (tc:definition->definition (car forms-list))))
		     defs)
		   (cdr forms-list)))
		 ((= 1 (length forms-list))
		  (proc pos
			(reverse defs)
			(tc:checked-form->exp (car forms-list))))
		 (else
		  (let* ((rev-forms (reverse forms-list))
			 (commands (reverse (cdr rev-forms)))
			 (command-exps (map tc:checked-form->exp commands))
			 (exp (tc:checked-form->exp (car rev-forms)))
			 (exp-pos (tc:scheme-exp->position exp))
			 (begin-pos (if (null? commands)
					exp-pos
					(tc:make-tree-position
					 (tc:position->start-line
					  (tc:scheme-form->position
					   (car commands)))
					 (tc:position->end-line exp-pos)
					 (tc:position->filename exp-pos))))
			 (sequence
			   (tc:begin-exp begin-pos command-exps exp)))
		    (proc pos (reverse defs) sequence)))))))
      (find-end-of-defs '() forms-list))))

;;; main public procedures

(define tc:parse-token-seq
  (lambda (start-state token-seq)
    (let ((answer (tc:parse-once start-state token-seq)))
      (cases tc:parser-answer-record answer
        (tc:parser-answer (position tree unparsed)
          (if (eq? (tc:scheme-token->class (tc:token-seq-head unparsed))
		   'end-marker)
              tree
	      (cases tc:scheme-token-record (tc:token-seq-head unparsed)
		(tc:scheme-token (position class data)
		   (tc:scheme-parser-error
		    position
		    "Tokens left over:"
		    (tc:token-class->string class)))
		(else (error "tc:parse-token-seq: tokens left over failed")))))
	(else (error "tc:parse-token-seq: bad answer" answer))))))

(define tc:parse-once
  (lambda (start-state token-seq)
    ((tc:goto-parser-state start-state) '() token-seq)))

(define tc:char-seq-program-parser
  (lambda (a-char-seq)
    (dynamic-wind
	tc:scheme-scanner-start-non-interactive-use!
	(lambda ()
	  (tc:parse-token-seq tc:parse-scheme-program
	    (tc:char-seq-scanner a-char-seq)))
  	tc:scheme-scanner-end-non-interactive-use!)))

(define tc:character-string-program-parser
  (lambda (a-string)
    (dynamic-wind
	tc:scheme-scanner-start-non-interactive-use!
	(lambda ()
	  (tc:parse-token-seq tc:parse-scheme-program
	    (tc:character-string-scanner a-string)))
  	tc:scheme-scanner-end-non-interactive-use!)))

(define tc:char-seq-form-parser
  (lambda (a-char-seq)
    (dynamic-wind
	tc:scheme-scanner-start-interactive-use!
	(lambda ()
	  (let* ((form (tc:parse-loads
		       (tc:parser-answer->tree
			(tc:parse-once tc:parse-scheme-form
			  (tc:char-seq-scanner a-char-seq)))))
		 (forms (tc:break-out-grouped-defines-in-forms (list form))))
	    (if (null? (cdr forms))
		(car forms)
		(tc:begin-form (tc:scheme-form->position form) forms))))
	tc:scheme-scanner-end-interactive-use!)))

(define tc:character-string-form-parser
  (lambda (a-string)
    (tc:parse-token-seq tc:parse-scheme-form
      (tc:character-string-scanner a-string))))

(define tc:character-string-exp-parser
  (lambda (a-string)
    (tc:parse-token-seq tc:parse-scheme-exp
      (tc:character-string-scanner a-string))))

(define tc:character-string-type-parser
  (lambda (a-string)
    (tc:parse-token-seq tc:parse-input-type-exp
      (tc:character-string-scanner a-string))))

;;; globals for giving better error messeges
(define *tc:token-classes-that-start-scheme-pattern*
  '(number boolean string character identifier lparen sharp-lparen))

(define *tc:token-classes-that-start-scheme-exp*
  '(number boolean string character identifier lparen
	   quote-char quasiquote-char))

(define *tc:token-classes-that-start-scheme-datum*
  '(number boolean string character identifier lparen
	   quote-char quasiquote-char unquote-char unquote-splicing-char
	   sharp-lparen))

(define *tc:token-classes-that-start-module-name*
  '(string identifier lparen))

(define *tc:token-classes-that-start-scheme-definition-guts*
  '( define begin define-syntax define-datatype deftype))

(define *tc:token-classes-that-start-module-body-form-guts*
  '( define begin define-syntax define-datatype deftype require defrep require-for-syntax))

;;; loads and files 

(define tc:parse-loads
  (lambda (form)
    (if (not *tc:parse-top-level-loads*)
	form
	(cases tc:scheme-form form
	  (tc:command (position exp)
            (if (not (tc:load-like-exp? exp))
		form
		(if (tc:unusable-load-like-exp? exp)
		    (begin
		      (display
		       (tc:position->error-string
			(tc:scheme-exp->position exp))
		       "Ignoring this load form; "
		       " it doesn't have a literal string argument"
		       (string-append tc:nl-indent "Ignored load form: ")
		       exp)
                      (newline)
		      form)
		    (tc:load-like-exp->filename
		     exp
		     (lambda (given-string path-to-file)
		       (let ((prog-tree
			      (tc:parse-scm-file
			       path-to-file position)))
			 (tc:load-command
			  position
			  (tc:varref->variable
			   (tc:procedure-call->operator exp))
			  given-string
			  prog-tree)))))))
	  (tc:begin-form (position forms)
	    (tc:begin-form position (map tc:parse-loads forms)))
	  (else form)))))

(deftype tc:load-like-exp? (-> (tc:scheme-exp) boolean))
(define tc:load-like-exp?
  (lambda (exp)
    (cases tc:scheme-exp exp
      (tc:procedure-call (position operator operands)
	  (cases tc:scheme-exp operator
	    (tc:varref (position variable)
	       (not (not
		     (memq variable
			   '(load load-quietly
				  load-from-lib load-quietly-from-lib)))))
	    (else #f)))
      (else #f))))

(deftype tc:unusable-load-like-exp? (-> (tc:scheme-exp) boolean))
(define tc:unusable-load-like-exp?
  (lambda (exp)
    ;; REQUIRES: (load-like-exp? exp) is true
    (cases tc:scheme-exp exp
      (tc:procedure-call (position operator operands)
	(or (null? operands)
	    (cases tc:scheme-exp (car operands)
	       (tc:self-evaluating (position datum)
		  (not (string? datum)))
	       (else #t))))
      (else (error "tc:unusable-load-like-exp passed a non procedure call!"
		   exp)))))

(deftype tc:load-like-exp->filename
  (forall (T) (-> (tc:scheme-exp (-> (string string) T)) T)))
(define tc:load-like-exp->filename
  (lambda (exp result-cont)
    ;; REQUIRES: (load-like-exp? exp) and not (tc:unusable-load-like-exp? exp)
    ;; ENSURES: result-cont is called with the name given in exp and
    ;; the actual path to the file.  The actual path to the file has
    ;; the library directory prepended if needed.  If the file
    ;; doesn't exist, then an error is signalled here, to prevent a
    ;; later error from the real load.
    (let ((operator (tc:varref->variable (tc:procedure-call->operator exp)))
	  (string-arg
	   (tc:self-evaluating->datum
	    (car (tc:procedure-call->operands exp)))))
      (let ((fname
	     (case operator
	       ((load load-quietly)
		(if (file-exists? string-arg)
		    string-arg
		    (let ((scmfile (tc:make-scm-file string-arg)))
		      (if (file-exists? scmfile)
			  scmfile
			  (string-append
			   (has-type-trusted string *cs342:libdir*)
			   (has-type-trusted string string-arg))))))
	       ((load-from-lib load-quietly-from-lib)
		(string-append
		 (has-type-trusted string *cs342:libdir*)
		 (has-type-trusted string string-arg)))
	       (else
		(error
		 "tc:load-like-exp->filename given unknown load operator!"
		 operator)))))
	(if (file-exists? fname)
	    (result-cont string-arg fname)
	    (if (not (tc:is-scm-file? fname))
		(let ((scmfile (tc:make-scm-file fname)))
		  (if (file-exists? scmfile)
		      (result-cont string-arg scmfile)
		      (error "Cannot find file: " scmfile)))
		(error "Cannot find file: " fname)))))))

(deftype tc:is-scm-file? (-> (string) boolean))
(define tc:is-scm-file?
  (lambda (fname)
    (equal? (tc:get-file-suffix fname) "scm")))

(define tc:make-scm-file
  (lambda (fname)
    (string-append fname ".scm")))

(deftype tc:get-file-prefix (-> (string) string))
(define tc:get-file-prefix
  (lambda (name)
    (let ((dot-pos (tc:last-dot-in-file-pos name (- (string-length name) 1))))
      (if (= -1 dot-pos)
	  name
	  (substring name 0 dot-pos)))))

(deftype tc:get-file-suffix (-> (string) string))
(define tc:get-file-suffix
  (lambda (name)
    (let* ((len (string-length name))
	   (dot-pos (tc:last-dot-in-file-pos name (- len 1))))
      (if (= -1 dot-pos)
	  ""
	  (substring name (+ 1 dot-pos) len)))))

(deftype tc:last-dot-in-file-pos (-> (string number) number))
(define tc:last-dot-in-file-pos
  (lambda (name pos)
    (if (< pos 0)
	-1
	(let ((current-char (string-ref name pos)))
	  (if (char=? #\. current-char)
	      pos
	      (if (or (char=? #\/ current-char) (char=? #\\ current-char))
		  -1
		  (tc:last-dot-in-file-pos name (- pos 1))))))))

(deftype tc:parse-scm-file
  (-> (string tc:position) tc:scheme-program))
(define tc:parse-scm-file
  (lambda (fname position)
    ;; EFFECT: manages the cache of parsed files.
    (maybe-test
     (tc:look-in-parsed-file-cache-for fname position)
     (lambda (program)
       ;;(displayln
       ;; "tc:parse-scm-file reusing parse of " fname)
       program)
     (lambda ()
       (let ((finished #f))
	 (dynamic-wind
	     (lambda ()
	       (tc:begin-parsing-file! fname))
	     (lambda ()
	       ;;(displayln "parsing " fname " ...")
	       (let ((program (tc:parse-possible-scm-file fname)))
		 (tc:done-parsing-file! fname program)
		 ;;(displayln ... done parsing " fname)
		 (set! finished #t)
		 program))
	     (lambda ()
	       (if (not finished)
		   (tc:remove-parsing-record-for! fname)))))))))
     
(deftype tc:parse-possible-scm-file (-> (string) tc:scheme-program))
(define tc:parse-possible-scm-file
  (lambda (fname)
    (if (not (file-exists? fname))
	(tc:scheme-program-normal
	 (tc:make-tree-position 0 0 fname)
	 '() '() '())
	(with-input-from-file fname
	  (lambda ()
	    (dynamic-wind
		(lambda ()
		  (tc:start-input-filename! fname))
		(lambda ()
		  (tc:char-seq-program-parser
		   (tc:stream->char-seq (tc:make-input-stream))))
		tc:end-input-filename!))))))


;;; grammar for programs --------------------------------------------

(define tc:parse-scheme-program
  (lambda (token)
    (tc:process-nt tc:parse-forms-to-end
      (tc:reduce 'scheme-program-normal))))

(define tc:parse-forms-to-end
  (lambda (token)
    ((tc:parse-loop* *tc:token-classes-that-start-scheme-exp* '(end-marker)
       (lambda (action) (tc:process-nt tc:parse-scheme-form action)))
     token)))

;;; grammar for forms *******************************************************

(define tc:parse-scheme-form
  (lambda (token)
    ((case (tc:scheme-token->class token)
       ((lparen) tc:seen-form-lparen)
       (else tc:parse-command))
     token)))

(define tc:parse-command
  (lambda (token)
    (tc:process-nt tc:parse-scheme-exp
      (tc:reduce 'command))))

(define tc:parse-command-guts
  (lambda (token)
    (tc:process-nt tc:parse-exp-guts
      (tc:reduce 'command))))

(define tc:seen-form-lparen
  (lambda (token)
    (tc:check/drop 'lparen
      (tc:goto-parser-state tc:parse-form-guts))))

(define tc:parse-form-guts
  (lambda (token)
    ((case (tc:scheme-token->class token)
       ((identifier)
	(case (tc:scheme-token->data token)
	  ((define) tc:seen-define)
	  ((begin) tc:seen-begin-form)
	  ((define-syntax) tc:seen-define-syntax)
	  ((define-datatype) tc:seen-define-datatype)
	  ((deftype) tc:seen-deftype)
	  ((module) tc:seen-module)
	  ((require) tc:seen-require-form)
	  (else tc:parse-command-guts)))
       (else tc:parse-command-guts))
     token)))

(define tc:seen-define
  (lambda (token)
    (tc:check/drop 'define
      (tc:process-nt tc:parse-define-guts
	(tc:reduce 'definition)))))

(define tc:seen-begin-form
  (lambda (token)
    (tc:check/drop 'begin
      (tc:process-nt tc:parse-form-list
	(tc:check/drop 'rparen
	  (tc:reduce 'begin-form))))))

(define tc:parse-form-list
  (lambda (token)
    ((tc:parse-loop* *tc:token-classes-that-start-scheme-exp* '(rparen)
       (lambda (action) (tc:process-nt tc:parse-scheme-form action)))
     token)))

(define tc:seen-define-datatype
  (lambda (token)
    (tc:check/drop 'define-datatype
      (tc:process-nt tc:parse-define-datatype-guts
	(tc:reduce 'definition)))))

(define tc:seen-deftype
  (lambda (token)
    (tc:check/drop 'deftype
      (tc:process-nt tc:parse-deftype-guts
	(tc:reduce 'definition)))))

(define tc:seen-define-syntax
  (lambda (token)
    (tc:check/drop 'define-syntax
      (tc:process-nt tc:parse-scheme-syntax-spec-guts ; eats the rparen
	(tc:reduce 'syntax-definition)))))

(define tc:seen-module
  (lambda (token)
    (tc:check/drop 'module
      (tc:check/shift 'identifier
        (tc:process-nt tc:parse-module-name
          (tc:process-nt tc:parse-provide
            (tc:process-nt tc:parse-module-body-form-list
              (tc:check/drop 'rparen
                (tc:reduce 'module)))))))))

(define tc:parse-provide
  (lambda (token)
    (tc:check/drop 'lparen
      (tc:check/drop 'provide
        (tc:process-nt tc:parse-no-dot-identifier-list
          (tc:check/drop 'rparen
            (tc:reduce 'provide)))))))

(define tc:parse-module-body-form-list
  (lambda (token)
    ((tc:parse-loop* '(lparen) '(rparen)
       (lambda (action) (tc:process-nt tc:parse-module-body-form action)))
     token)))

(define tc:parse-module-body-form
  (lambda (token)
    (tc:check/drop 'lparen
      (tc:goto-parser-state tc:parse-module-body-form-guts))))

(define tc:parse-module-body-form-guts
  (lambda (token)
    ((case (tc:scheme-token->class token)
       ((identifier)
	(case (tc:scheme-token->data token)
	  ((define) tc:seen-define-definition)
	  ((begin) tc:seen-begin-definition)
	  ((define-syntax) tc:seen-define-syntax-definition)
	  ((define-datatype) tc:seen-define-datatype-definition)
	  ((deftype) tc:seen-deftype-definition)
	  ((require) tc:seen-require)
	  ((defrep) tc:seen-defrep)
	  ((require-for-syntax) tc:seen-require-for-syntax)
	  (else (tc:unexpected-error
		 *tc:token-classes-that-start-module-body-form-guts*))))
       (else (tc:unexpected-error
	      *tc:token-classes-that-start-module-body-form-guts*)))
     token)))

(define tc:seen-define-definition
  (lambda (token)
    (tc:check/drop 'define
      (tc:goto-parser-state tc:parse-define-guts))))

(define tc:seen-begin-definition
  (lambda (token)
    (tc:check/drop 'begin
      (tc:process-nt tc:parse-definition-list
     	(tc:check/drop 'rparen
	  (tc:reduce 'grouped-define))))))

(define tc:seen-define-syntax-definition
  (lambda (token)
    (tc:check/drop 'define-syntax
      (tc:goto-parser-state tc:parse-scheme-syntax-spec-guts))))

(define tc:seen-define-datatype-definition
  (lambda (token)
    (tc:check/drop 'define-datatype
      (tc:goto-parser-state tc:parse-define-datatype-guts))))

(define tc:seen-deftype-definition
  (lambda (token)
    (tc:check/drop 'deftype
      (tc:goto-parser-state tc:parse-deftype-guts))))

(define tc:parse-definition-list
  (lambda (token)
    ((tc:parse-loop* '(lparen) '(rparen)
       (lambda (action) (tc:process-nt tc:parse-definition action)))
     token)))

(define tc:parse-definition
  (lambda (token)
    (tc:check/drop 'lparen
      (tc:goto-parser-state tc:parse-definition-guts))))

(define tc:parse-definition-guts
  (lambda (token)
    ((case (tc:scheme-token->class token)
       ((identifier)
	(case (tc:scheme-token->data token)
	  ((define) tc:seen-define)
	  ((begin) tc:seen-begin-definition)
	  ((define-syntax) tc:seen-define-syntax)
	  ((define-datatype) tc:seen-define-datatype)
	  ((deftype) tc:seen-deftype)
	  (else (tc:unexpected-error
		 *tc:token-classes-that-start-scheme-definition-guts*))))
       (else (tc:unexpected-error
	      *tc:token-classes-that-start-scheme-definition-guts*)))
     token)))

(define tc:seen-defrep
  (lambda (token)
    (tc:check/drop 'defrep
      (tc:process-nt tc:parse-abs-conc-pair-list
        (tc:check/drop 'rparen
          (tc:reduce 'defrep))))))

(define tc:parse-abs-conc-pair-list
  (lambda (token)
    ((tc:parse-loop+ '(lparen) '(rparen)
       (lambda (action) (tc:process-nt tc:parse-abs-conc-pair action)))
     token)))

(define tc:parse-abs-conc-pair
  (lambda (token)
    (tc:check/drop 'lparen
      (tc:goto-parser-state tc:parse-abs-conc-pair-guts))))

(define tc:parse-abs-conc-pair-guts
  (lambda (token)
    ((case (tc:scheme-token->class token)
       ((lparen) tc:seen-abs-conc-pair-lparen-or-id)
       ((identifier)
	(case (tc:scheme-token->data token)
	  ((forall) tc:seen-abs-conc-forall)
	  (else tc:seen-abs-conc-pair-lparen-or-id)))
       (else (tc:unexpected-error '(lparen forall identifier))))
     token)))

(define tc:seen-abs-conc-pair-lparen-or-id
  (lambda (token)
    (tc:process-nt tc:parse-input-type-exp
      (tc:process-nt tc:parse-input-type-exp
	 (tc:check/drop 'rparen
	   (tc:reduce 'abs-conc-pair-simple))))))

(define tc:seen-abs-conc-forall
  (lambda (token)
    (tc:check/drop 'forall
      (tc:check/drop 'lparen
        (tc:process-nt tc:parse-no-dot-identifier-list
	  (tc:check/drop 'rparen
            (tc:process-nt tc:parse-input-type-exp
	      (tc:process-nt tc:parse-input-type-exp
	        (tc:check/drop 'rparen
	          (tc:reduce 'abs-conc-pair-polymorphic))))))))))

(define tc:seen-require-for-syntax
  (lambda (token)
    (tc:check/drop 'require-for-syntax
       (tc:process-nt tc:parse-module-name-list
	  (tc:check/drop 'rparen
	     (tc:reduce 'require-for-syntax))))))

(define tc:seen-require-form
  (lambda (token)
    (tc:check/drop 'require
      (tc:process-nt tc:parse-module-name-list
        (tc:check/drop 'rparen
          (tc:reduce 'require-form))))))

(define tc:seen-require
  (lambda (token)
    (tc:check/drop 'require
      (tc:process-nt tc:parse-module-name-list
        (tc:check/drop 'rparen
          (tc:reduce 'require))))))

(define tc:parse-module-name-list
  (lambda (token)
    ((tc:parse-loop* *tc:token-classes-that-start-module-name* '(rparen)
       (lambda (action) (tc:process-nt tc:parse-module-name action)))
     token)))

(define tc:parse-module-name
  (lambda (token)
    ((case (tc:scheme-token->class token)
       ((lparen) tc:seen-module-name-lparen)
       ((identifier) tc:seen-module-name-identifier)
       ((string) tc:seen-module-name-string)
       (else (tc:unexpected-error '(lparen identifier string))))
     token)))

(define tc:seen-module-name-identifier
  (lambda (token)
    (tc:check/shift 'identifier
      (tc:reduce 'module-id))))

(define tc:seen-module-name-string
  (lambda (token)
    (tc:check/shift 'string
      (tc:reduce 'unix-relative-path-string))))

(define tc:seen-module-name-lparen
  (lambda (token)
    (tc:check/drop 'lparen
      (tc:goto-parser-state tc:parse-module-name-guts))))

(define tc:parse-module-name-guts
  (lambda (token)
    ((case (tc:scheme-token->class token)
       ((identifier)
	(case (tc:scheme-token->data token)
	  ((file) tc:seen-module-name-file)
	  ((lib) tc:seen-module-name-lib)
	  (else (tc:unexpected-error '(file lib)))))
       (else (tc:unexpected-error '(file lib))))
     token)))

(define tc:seen-module-name-file
  (lambda (token)
    (tc:check/drop 'file
      (tc:check/shift 'string
        (tc:check/drop 'rparen
          (tc:reduce 'file))))))

(define tc:seen-module-name-lib
  (lambda (token)
    (tc:check/drop 'lib
      (tc:check/shift 'string
	(tc:process-nt tc:parse-string-list
          (tc:check/drop 'rparen
            (tc:reduce 'lib)))))))

(define tc:parse-string-list
  (lambda (token)
    ((tc:parse-loop* '(string) '(rparen)
       (lambda (action) (tc:check/shift 'string action)))
     token)))

;;; definitions

(define tc:parse-define-guts
  (lambda (token)
    ((case (tc:scheme-token->class token)
      ((identifier) tc:parse-simple-define)
      ((lparen) tc:parse-procedure-define)
      (else (tc:unexpected-error '(lparen identifier))))
     token)))

(define tc:parse-simple-define
  (lambda (token)
    (tc:check/shift 'identifier
      (tc:process-nt tc:parse-scheme-exp
	(tc:check/drop 'rparen
	  (tc:reduce 'simple-define))))))

(define tc:parse-procedure-define
  (lambda (token)
    (tc:check/drop 'lparen
      (tc:check/shift 'identifier
	(tc:process-nt tc:parse-formals-guts
	  (tc:check/drop 'rparen
	    (tc:process-nt tc:parse-body
	      (tc:check/drop 'rparen
	        (tc:reduce 'procedure-define)))))))))

(define tc:parse-define-datatype-guts
  (lambda (token)
    (tc:check/shift 'identifier
      (tc:check/shift 'identifier
        (tc:process-nt tc:parse-variant-decl-list
	  (tc:check/drop 'rparen
	    (tc:reduce 'define-datatype)))))))

(define tc:parse-deftype-guts
  (lambda (token)
    (tc:check/shift 'identifier
      (tc:process-nt tc:parse-input-type-exp
	(tc:check/drop 'rparen
	  (tc:reduce 'deftype))))))

;;; formals

(define tc:parse-formals-guts
  (lambda (token)
    (tc:process-nt tc:parse-identifier-list
      (tc:goto-parser-state tc:parse-optional-vararg-formal))))

(define tc:parse-identifier-list
  (lambda (token)
    ((tc:parse-loop* '(identifier) '(dot rparen)
       (lambda (action) (tc:check/shift 'identifier action)))
     token)))

(define tc:parse-optional-vararg-formal
  (lambda (token)
    ((case (tc:scheme-token->class token)
      ((rparen) tc:parse-end-of-simple-formals)
      ((dot) tc:parse-optional-formal)
      (else (tc:unexpected-error '(rparen dot))))
     token)))

(define tc:parse-end-of-simple-formals
  (lambda (token)
    (tc:reduce 'normal-formals)))

(define tc:parse-optional-formal
  (lambda (token)
    (tc:check/drop 'dot
      (tc:check/shift 'identifier
	(tc:reduce 'some-normal-formals)))))

;;; body
(define tc:parse-body
  (lambda (token)
    (tc:process-nt tc:parse-definition-command-list
      (tc:reduce 'body))))

(define tc:parse-definition-command-list
  (lambda (token)
    ((tc:parse-loop* *tc:token-classes-that-start-scheme-exp* '(rparen)
       (lambda (action) (tc:process-nt tc:parse-definition-or-command action)))
     token)))

(define tc:parse-definition-or-command
  (lambda (token)
    ((case (tc:scheme-token->class token)
       ((lparen) tc:seen-def-comm-lparen)
       (else tc:parse-command))
     token)))

(define tc:seen-def-comm-lparen
  (lambda (token)
    (tc:check/drop 'lparen
      (tc:goto-parser-state tc:parse-def-comm-guts))))

(define tc:parse-def-comm-guts
  (lambda (token)
    ((case (tc:scheme-token->class token)
       ((identifier)
	(case (tc:scheme-token->data token)
	  ((define) tc:seen-define)
	  ((begin) tc:seen-begin-def-comm)
	  ((define-datatype) tc:seen-define-datatype)
	  ((deftype) tc:seen-deftype)
	  (else tc:parse-command-guts)))
       (else tc:parse-command-guts))
     token)))

(define tc:seen-begin-def-comm
  (lambda (token)
    (tc:check/drop 'begin
      (tc:process-nt tc:parse-definition-command-list
	(tc:check/drop 'rparen
	  (tc:reduce 'begin-def-comm))))))

;;; variant-decls
(define tc:parse-variant-decl-list
  (lambda (token)
    ((tc:parse-loop+ '(lparen) '(rparen)
       (lambda (action) (tc:process-nt tc:parse-variant-decl action)))
     token)))

(define tc:parse-variant-decl
  (lambda (token)
    (tc:check/drop 'lparen
      (tc:check/shift 'identifier
	(tc:process-nt tc:parse-field-pred-bindings
	  (tc:check/drop 'rparen
	    (tc:reduce 'variant-decl)))))))

;;; field-pred bindings
(define tc:parse-field-pred-bindings
  (lambda (token)
    ((tc:parse-loop* '(lparen) '(rparen)
       (lambda (action) (tc:process-nt tc:parse-field-pred-binding action)))
     token)))

(define tc:parse-field-pred-binding
  (lambda (token)
    (tc:check/drop 'lparen
      (tc:check/shift 'identifier
	(tc:process-nt tc:parse-scheme-exp
	  (tc:check/drop 'rparen
	    (tc:reduce 'field-pred-binding)))))))

;;; scheme syntax specs
(define tc:parse-scheme-syntax-spec
  (lambda (token)
    (tc:check/drop 'lparen
      (tc:goto-parser-state tc:parse-scheme-syntax-spec-guts))))

(define tc:parse-scheme-syntax-spec-guts
  (lambda (token)
    (tc:check/shift 'identifier
      (tc:process-nt tc:parse-scheme-transformer-spec
        (tc:check/drop 'rparen
	  (tc:reduce 'scheme-syntax-spec))))))

;;; scheme transformer specs
(define tc:parse-scheme-transformer-spec
  (lambda (token)
    (tc:check/drop 'lparen
      (tc:check/drop 'syntax-rules
	(tc:check/drop 'lparen
          (tc:process-nt tc:parse-no-dot-identifier-list
            (tc:check/drop 'rparen
	      (tc:process-nt tc:parse-syntax-rule-list
                (tc:check/drop 'rparen
		  (tc:reduce 'scheme-transformer-spec))))))))))

(define tc:parse-syntax-rule-list
  (lambda (token)
    ((tc:parse-loop* '(lparen) '(rparen)
       (lambda (action) (tc:process-nt tc:parse-syntax-rule action)))
     token)))

;;; syntax rules
(define tc:parse-syntax-rule
  (lambda (token)
    (tc:check/drop 'lparen
      (tc:process-nt tc:parse-scheme-syntax-pattern
	(tc:process-nt tc:parse-scheme-syntax-template
	  (tc:check/drop 'rparen
	    (tc:reduce 'scheme-syntax-rule)))))))

;;; syntax patterns
(define tc:parse-scheme-syntax-pattern
  (lambda (token)
    ((case (tc:scheme-token->class token)
      ((identifier) tc:seen-identifier-pattern)
      ((number) tc:seen-number-pattern)
      ((boolean) tc:seen-boolean-pattern)
      ((string) tc:seen-string-pattern)
      ((character) tc:seen-character-pattern)
      ((lparen) tc:seen-pattern-lparen)
      ((sharp-lparen) tc:seen-pattern-sharp-lparen)
      (else (tc:unexpected-error
	     *tc:token-classes-that-start-scheme-pattern*)))
     token)))

(define tc:seen-identifier-pattern
  (lambda (token)
    (tc:check/shift 'identifier
      (tc:reduce 'syntax-pattern-identifier))))

(define tc:seen-number-pattern
  (lambda (token)
    (tc:check/shift 'number
      (tc:reduce 'syntax-pattern-datum))))

(define tc:seen-boolean-pattern
  (lambda (token)
    (tc:check/shift 'boolean
      (tc:reduce 'syntax-pattern-datum))))

(define tc:seen-string-pattern
  (lambda (token)
    (tc:check/shift 'string
      (tc:reduce 'syntax-pattern-datum))))

(define tc:seen-character-pattern
  (lambda (token)
    (tc:check/shift 'character
      (tc:reduce 'syntax-pattern-datum))))

(define tc:seen-pattern-lparen
  (lambda (token)
    (tc:check/drop 'lparen
      (tc:goto-parser-state tc:parse-list-pattern-guts))))

(define tc:parse-list-pattern-guts
  (lambda (token)
    (tc:process-nt tc:parse-scheme-syntax-pattern-list
      (tc:goto-parser-state tc:parse-optional-dotted-pattern-list))))

(define tc:parse-scheme-syntax-pattern-list
  (lambda (token)
    ((tc:parse-loop* *tc:token-classes-that-start-scheme-pattern* '(dot rparen)
       (lambda (action) (tc:process-nt tc:parse-scheme-syntax-pattern action)))
     token)))

(define tc:parse-optional-dotted-pattern-list
  (lambda (token)
    ((case (tc:scheme-token->class token)
      ((rparen) tc:parse-end-of-normal-pattern-list)
      ((dot) tc:parse-optional-dotted-pattern-item)
      (else (tc:unexpected-error '(rparen dot))))
     token)))

(define tc:parse-end-of-normal-pattern-list
  (lambda (token)
    (tc:check/drop 'rparen
      (tc:reduce 'syntax-pattern-normal-list))))

(define tc:parse-optional-dotted-pattern-item
  (lambda (token)
    (tc:check/drop 'dot
      (tc:process-nt tc:parse-scheme-syntax-pattern
	(tc:check/drop 'rparen
	  (tc:reduce 'syntax-pattern-dotted-list))))))

(define tc:seen-pattern-sharp-lparen
  (lambda (token)
    (tc:check/drop 'sharp-lparen
      (tc:process-nt tc:parse-pattern-vector-elements
	(tc:check/drop 'rparen
	  (tc:reduce 'syntax-pattern-normal-vector))))))

(define tc:parse-pattern-vector-elements
  (lambda (token)
    ((tc:parse-loop* *tc:token-classes-that-start-scheme-pattern* '(rparen)
       (lambda (action) (tc:process-nt tc:parse-scheme-syntax-pattern action)))
     token)))

;;; syntax templates
(define tc:parse-scheme-syntax-template
  (lambda (token)
    ((case (tc:scheme-token->class token)
      ((identifier) tc:seen-identifier-template)
      ((number) tc:seen-number-template)
      ((boolean) tc:seen-boolean-template)
      ((string) tc:seen-string-template)
      ((character) tc:seen-character-template)
      ((lparen) tc:seen-template-lparen)
      ((sharp-lparen) tc:seen-template-sharp-lparen)
      (else (tc:unexpected-error
	     *tc:token-classes-that-start-scheme-pattern*)))
     token)))

(define tc:seen-identifier-template
  (lambda (token)
    (tc:check/shift 'identifier
      (tc:reduce 'syntax-template-identifier))))

(define tc:seen-number-template
  (lambda (token)
    (tc:check/shift 'number
      (tc:reduce 'syntax-template-datum))))

(define tc:seen-boolean-template
  (lambda (token)
    (tc:check/shift 'boolean
      (tc:reduce 'syntax-template-datum))))

(define tc:seen-string-template
  (lambda (token)
    (tc:check/shift 'string
      (tc:reduce 'syntax-template-datum))))

(define tc:seen-character-template
  (lambda (token)
    (tc:check/shift 'character
      (tc:reduce 'syntax-template-datum))))

(define tc:seen-template-lparen
  (lambda (token)
    (tc:check/drop 'lparen
      (tc:goto-parser-state tc:parse-list-template-guts))))

(define tc:parse-list-template-guts
  (lambda (token)
    (tc:process-nt tc:parse-scheme-syntax-template-element-list
      (tc:goto-parser-state tc:parse-optional-dotted-template-element-list))))

(define tc:parse-scheme-syntax-template-element-list
  (lambda (token)
    ((tc:parse-loop* *tc:token-classes-that-start-scheme-pattern* '(dot rparen)
       (lambda (action)
	 (tc:process-nt tc:parse-scheme-syntax-template-element action)))
     token)))

(define tc:parse-optional-dotted-template-element-list
  (lambda (token)
    ((case (tc:scheme-token->class token)
      ((rparen) tc:parse-end-of-normal-template-element-list)
      ((dot) tc:parse-optional-dotted-template-item)
      (else (tc:unexpected-error '(rparen dot))))
     token)))

(define tc:parse-end-of-normal-template-element-list
  (lambda (token)
    (tc:check/drop 'rparen
      (tc:reduce 'syntax-template-normal-list))))

(define tc:parse-optional-dotted-template-item
  (lambda (token)
    (tc:check/drop 'dot
      (tc:process-nt tc:parse-scheme-syntax-template
	(tc:check/drop 'rparen
	  (tc:reduce 'syntax-template-dotted-list))))))

(define tc:seen-template-sharp-lparen
  (lambda (token)
    (tc:check/drop 'sharp-lparen
      (tc:process-nt tc:parse-template-vector-elements
	(tc:check/drop 'rparen
	  (tc:reduce 'syntax-template-normal-vector))))))

(define tc:parse-template-vector-elements
  (lambda (token)
    ((tc:parse-loop* *tc:token-classes-that-start-scheme-pattern* '(rparen)
       (lambda (action) (tc:process-nt
			 tc:parse-scheme-syntax-template-element action)))
     token)))

;;; syntax template elements
(define tc:parse-scheme-syntax-template-element
  (lambda (token)
    (tc:process-nt tc:parse-scheme-syntax-template
      (tc:goto-parser-state tc:parse-optional-ellipsis-in-template))))

(define tc:parse-optional-ellipsis-in-template
  (lambda (token)
    ((case (tc:scheme-token->class token)
      ((identifier)
       (case (tc:scheme-token->data token)
	 ((...) tc:seen-ellipsis-template-element)
	 (else tc:seen-no-ellipsis-template-element)))
      (else tc:seen-no-ellipsis-template-element))
     token)))

(define tc:seen-no-ellipsis-template-element
  (lambda (token)
    (tc:reduce 'scheme-syntax-template-element-normal)))

(define tc:seen-ellipsis-template-element
  (lambda (token)
    (tc:check/drop '...
      (tc:reduce 'scheme-syntax-template-element-ellipsis))))

;;; grammar for expressions *************************************************

(define tc:parse-scheme-exp
  (lambda (token)
    ((case (tc:scheme-token->class token)
      ((identifier) tc:seen-identifier)
      ((number) tc:seen-number)
      ((boolean) tc:seen-boolean)
      ((string) tc:seen-string)
      ((character) tc:seen-character)
      ((lparen) tc:seen-exp-lparen)
      ((quote-char) tc:seen-quote-char)
      ((quasiquote-char) tc:seen-quasiquote-char)
      (else (tc:unexpected-error
	     *tc:token-classes-that-start-scheme-exp*)))
     token)))

(define tc:seen-identifier
  (lambda (token)
    (tc:check/shift 'identifier (tc:reduce 'varref))))

(define tc:seen-number
  (lambda (token)
    (tc:check/shift 'number (tc:reduce 'self-evaluating))))

(define tc:seen-boolean
  (lambda (token)
    (tc:check/shift 'boolean (tc:reduce 'self-evaluating))))

(define tc:seen-string
  (lambda (token)
    (tc:check/shift 'string (tc:reduce 'self-evaluating))))

(define tc:seen-character
  (lambda (token)
    (tc:check/shift 'character (tc:reduce 'self-evaluating))))

(define tc:seen-quote-char
  (lambda (token)
    (tc:check/drop 'quote-char
      (tc:process-nt tc:parse-scheme-datum
        (tc:reduce 'quotation)))))

(define tc:seen-quasiquote-char
  (lambda (token)
    (tc:check/drop 'quasiquote-char
      (tc:process-qq-nt! tc:parse-scheme-quasiquoted-datum
        (tc:reduce 'quasiquotation)))))

(define tc:seen-quote
  (lambda (token)
    (tc:check/drop 'quote
      (tc:process-nt tc:parse-scheme-datum
        (tc:check/drop 'rparen
          (tc:reduce 'quotation))))))

(define tc:seen-quasiquote
  (lambda (token)
    (tc:check/drop 'quasiquote
      (tc:process-qq-nt! tc:parse-scheme-quasiquoted-datum
        (tc:check/drop 'rparen
          (tc:reduce 'quasiquotation))))))

(define tc:seen-exp-lparen
  (lambda (token)
    (tc:check/drop 'lparen
      (tc:goto-parser-state tc:parse-exp-guts))))

(define tc:parse-exp-guts
  (lambda (token)
    ((case (tc:scheme-token->class token)
      ((rparen) tc:seen-empty-list)
      ((identifier)
       (case (tc:scheme-token->data token)
	 ((quote) tc:seen-quote)
	 ((lambda) tc:seen-lambda)
	 ((if) tc:seen-if)
	 ((set!) tc:seen-set!)
	 ((let-syntax) tc:seen-let-syntax)
	 ((letrec-syntax) tc:seen-letrec-syntax)
	 ((cond) tc:seen-cond)
	 ((case) tc:seen-case)
	 ((and) tc:seen-and)
	 ((or) tc:seen-or)
	 ((begin) tc:seen-begin)
	 ((let) tc:seen-let)
	 ((let*) tc:seen-let*)
	 ((letrec) tc:seen-letrec)
	 ((do) tc:seen-do)
	 ((delay) tc:seen-delay)
	 ((quasiquote) tc:seen-quasiquote)
	 ((cases) tc:seen-cases)
	 ((has-type) tc:seen-has-type)
	 ((has-type-trusted) tc:seen-has-type-trusted)
	 ((test-type) tc:seen-test-type)
	 ((->) tc:seen-arrow-exp)
	 ((forall) tc:seen-forall-exp)
	 (else tc:seen-procedure-call)))
      (else tc:seen-procedure-call))
     token)))

(define tc:seen-empty-list
  (lambda (token)
    (tc:check/drop 'rparen
      (tc:reduce 'empty-list))))

(define tc:seen-lambda
  (lambda (token)
    (tc:check/drop 'lambda
      (tc:goto-parser-state tc:parse-lambda-formals))))

(define tc:parse-lambda-formals
  (lambda (token)
    ((case (tc:scheme-token->class token)
       ((lparen) tc:seen-lambda-lparen)
       ((identifier) tc:seen-lambda-unrestricted-formals)
       (else (tc:unexpected-error '(lparen identifier))))
     token)))

(define tc:seen-lambda-lparen
  (lambda (token)
    (tc:check/drop 'lparen
      (tc:process-nt tc:parse-formals-guts
	(tc:check/drop 'rparen
	  (tc:process-nt tc:parse-body
	    (tc:check/drop 'rparen
	      (tc:reduce 'lambda-exp))))))))

(define tc:seen-lambda-unrestricted-formals
  (lambda (token)
    (tc:process-nt tc:parse-unrestricted-formals
      (tc:process-nt tc:parse-body
	(tc:check/drop 'rparen
	  (tc:reduce 'lambda-exp))))))

(define tc:parse-unrestricted-formals
  (lambda (token)
    (tc:check/shift 'identifier
      (tc:reduce 'unrestricted-formals))))

(define tc:seen-if
  (lambda (token)
    (tc:check/drop 'if
      (tc:process-nt tc:parse-scheme-exp
	(tc:process-nt tc:parse-scheme-exp
	 (tc:goto-parser-state tc:parse-optional-if-else))))))

(define tc:parse-optional-if-else
  (lambda (token)
    ((case (tc:scheme-token->class token)
       ((rparen) tc:parse-if-no-else)
       (else tc:parse-if-else))
     token)))

(define tc:parse-if-else
  (lambda (token)
    (tc:process-nt tc:parse-scheme-exp
      (tc:check/drop 'rparen
        (tc:reduce 'conditional-exp)))))

(define tc:parse-if-no-else
  (lambda (token)
    (tc:check/drop 'rparen
      (tc:reduce 'conditional-stmt))))

(define tc:seen-set!
  (lambda (token)
    (tc:check/drop 'set!
      (tc:check/shift 'identifier
	(tc:process-nt tc:parse-scheme-exp
	  (tc:check/drop 'rparen
	    (tc:reduce 'assignment)))))))

(define tc:seen-let-syntax
  (lambda (token)
    (tc:check/drop 'let-syntax
      (tc:check/drop 'lparen
        (tc:process-nt tc:parse-scheme-syntax-spec-list
	  (tc:check/drop 'rparen
            (tc:process-nt tc:parse-body
	      (tc:check/drop 'rparen
                (tc:reduce 'let-syntax-exp)))))))))

(define tc:parse-scheme-syntax-spec-list
  (lambda (token)
    ((tc:parse-loop* '(lparen) '(rparen)
       (lambda (action) (tc:process-nt tc:parse-scheme-syntax-spec action)))
     token)))

(define tc:seen-letrec-syntax
  (lambda (token)
    (tc:check/drop 'letrec-syntax
      (tc:check/drop 'lparen
        (tc:process-nt tc:parse-scheme-syntax-spec-list
	  (tc:check/drop 'rparen
            (tc:process-nt tc:parse-body
	      (tc:check/drop 'rparen
                (tc:reduce 'letrec-syntax-exp)))))))))

(define tc:seen-begin
  (lambda (token)
    (tc:check/drop 'begin
      (tc:process-nt tc:parse-nonempty-exp-list
        (tc:check/drop 'rparen
	   (tc:reduce 'begin-exp))))))

(define tc:parse-nonempty-exp-list
  (lambda (token)
    ((tc:parse-loop+ *tc:token-classes-that-start-scheme-exp* '(rparen)
       (lambda (action) (tc:process-nt tc:parse-scheme-exp action)))
     token)))

(define tc:seen-let
  (lambda (token)
    (tc:check/drop 'let
      (tc:goto-parser-state tc:parse-let-guts))))

(define tc:parse-let-guts
  (lambda (token)
    ((case (tc:scheme-token->class token)
       ((lparen) tc:seen-let-lparen)
       ((identifier) tc:seen-named-let)
       (else (tc:unexpected-error '(lparen identifier))))
     token)))

(define tc:seen-let-lparen
  (lambda (token)
    (tc:check/drop 'lparen
      (tc:process-nt tc:parse-scheme-binding-specs
	(tc:check/drop 'rparen
          (tc:process-nt tc:parse-body
	    (tc:check/drop 'rparen
              (tc:reduce 'let-exp))))))))

(define tc:parse-scheme-binding-specs
  (lambda (token)
    ((tc:parse-loop+ '(lparen) '(rparen)
       (lambda (action) (tc:process-nt tc:parse-scheme-binding-spec action)))
     token)))

(define tc:parse-scheme-binding-spec
  (lambda (token)
    (tc:check/drop 'lparen
      (tc:check/shift 'identifier
	(tc:process-nt tc:parse-scheme-exp
	  (tc:check/drop 'rparen
             (tc:reduce 'scheme-binding-spec)))))))

(define tc:seen-named-let
  (lambda (token)
    (tc:check/shift 'identifier
      (tc:check/drop 'lparen
        (tc:process-nt tc:parse-optional-scheme-binding-specs
	  (tc:check/drop 'rparen
            (tc:process-nt tc:parse-body
	      (tc:check/drop 'rparen
                (tc:reduce 'named-let-exp)))))))))

(define tc:parse-optional-scheme-binding-specs
  (lambda (token)
    ((tc:parse-loop* '(lparen) '(rparen)
       (lambda (action) (tc:process-nt tc:parse-scheme-binding-spec action)))
     token)))

(define tc:seen-let*
  (lambda (token)
    (tc:check/drop 'let*
      (tc:check/drop 'lparen
        (tc:process-nt tc:parse-scheme-binding-specs
	  (tc:check/drop 'rparen
            (tc:process-nt tc:parse-body
	      (tc:check/drop 'rparen
                (tc:reduce 'let*-exp)))))))))

(define tc:seen-letrec
  (lambda (token)
    (tc:check/drop 'letrec
      (tc:check/drop 'lparen
        (tc:process-nt tc:parse-scheme-binding-specs
	  (tc:check/drop 'rparen
            (tc:process-nt tc:parse-body
	      (tc:check/drop 'rparen
                (tc:reduce 'letrec-exp)))))))))

(define tc:seen-procedure-call
  (lambda (token)
    (tc:process-nt tc:parse-nonempty-exp-list
      (tc:check/drop 'rparen
	(tc:reduce 'procedure-call)))))

(define tc:seen-cond
  (lambda (token)
    (tc:check/drop 'cond
      (tc:process-nt tc:parse-cond-clause-list
	(tc:check/drop 'rparen
	  (tc:reduce 'cond))))))

(define tc:parse-cond-clause-list
  (lambda (token)
    ((tc:parse-loop+ '(lparen) '(rparen)
       (lambda (action) (tc:process-nt tc:parse-cond-clause action)))
     token)))

(define tc:seen-case
  (lambda (token)
    (tc:check/drop 'case
      (tc:process-nt tc:parse-scheme-exp
        (tc:process-nt tc:parse-case-clause-list
	  (tc:check/drop 'rparen
	    (tc:reduce 'case)))))))

(define tc:parse-case-clause-list
  (lambda (token)
    ((tc:parse-loop+ '(lparen) '(rparen)
       (lambda (action) (tc:process-nt tc:parse-case-clause action)))
     token)))

(define tc:seen-and
  (lambda (token)
    (tc:check/drop 'and
      (tc:process-nt tc:parse-exp-list
        (tc:check/drop 'rparen
	  (tc:reduce 'and-exp))))))

(define tc:seen-or
  (lambda (token)
    (tc:check/drop 'or
      (tc:process-nt tc:parse-exp-list
        (tc:check/drop 'rparen
	  (tc:reduce 'or-exp))))))

(define tc:parse-exp-list
  (lambda (token)
    ((tc:parse-loop* *tc:token-classes-that-start-scheme-exp* '(rparen)
       (lambda (action) (tc:process-nt tc:parse-scheme-exp action)))
     token)))

(define tc:seen-do
  (lambda (token)
    (tc:check/drop 'do
      (tc:check/drop 'lparen
        (tc:process-nt tc:parse-iteration-spec-list
          (tc:check/drop 'rparen
            (tc:check/drop 'lparen
              (tc:process-nt tc:parse-scheme-exp
		(tc:goto-parser-state tc:parse-optional-do-result)))))))))

(define tc:parse-iteration-spec-list
  (lambda (token)
    ((tc:parse-loop* '(lparen) '(rparen)
      (lambda (action) (tc:process-nt tc:parse-scheme-iteration-spec action)))
     token)))

(define tc:parse-optional-do-result
  (lambda (token)
    ((case (tc:scheme-token->class token)
       ((rparen) tc:parse-rest-of-do-stmt)
       (else tc:parse-rest-of-do-exp))
     token)))

(define tc:parse-rest-of-do-exp
  (lambda (token)
    (tc:process-nt tc:parse-sequence
      (tc:check/drop 'rparen
	(tc:process-nt tc:parse-exp-list
	  (tc:check/drop 'rparen
	    (tc:reduce 'do-exp)))))))

(define tc:parse-rest-of-do-stmt
  (lambda (token)
     (tc:check/drop 'rparen
       (tc:process-nt tc:parse-exp-list
	 (tc:check/drop 'rparen
	   (tc:reduce 'do-stmt))))))

(define tc:seen-delay
  (lambda (token)
    (tc:check/drop 'delay
      (tc:process-nt tc:parse-scheme-exp
        (tc:check/drop 'rparen
	  (tc:reduce 'delay-exp))))))

(define tc:seen-cases
  (lambda (token)
    (tc:check/drop 'cases
      (tc:check/shift 'identifier
        (tc:process-nt tc:parse-scheme-exp
          (tc:process-nt tc:parse-cases-clause-list
	    (tc:check/drop 'rparen
	      (tc:reduce 'cases))))))))

(define tc:parse-cases-clause-list
  (lambda (token)
    ((tc:parse-loop+ '(lparen) '(rparen)
       (lambda (action) (tc:process-nt tc:parse-cases-clause action)))
     token)))

(define tc:seen-has-type
  (lambda (token)
    (tc:check/drop 'has-type
      (tc:process-nt tc:parse-input-type-exp
        (tc:process-nt tc:parse-scheme-exp
          (tc:check/drop 'rparen
	    (tc:reduce 'has-type-exp)))))))

(define tc:seen-has-type-trusted
  (lambda (token)
    (tc:check/drop 'has-type-trusted
      (tc:process-nt tc:parse-input-type-exp
        (tc:process-nt tc:parse-scheme-exp
          (tc:check/drop 'rparen
	    (tc:reduce 'has-type-trusted-exp)))))))

(define tc:seen-test-type
  (lambda (token)
    (tc:check/drop 'test-type
      (tc:process-nt tc:parse-scheme-exp
        (tc:process-nt tc:parse-scheme-exp
          (tc:check/drop 'rparen
	    (tc:reduce 'test-type-exp)))))))

(define tc:seen-arrow-exp
  (lambda (token)
    (tc:check/drop '->
      (tc:check/drop 'lparen
	(tc:process-nt tc:parse-exp-list
	  (tc:check/drop 'rparen
	    (tc:process-nt tc:parse-scheme-exp
              (tc:check/drop 'rparen
	        (tc:reduce 'arrow-exp)))))))))

(define tc:seen-forall-exp
  (lambda (token)
    (tc:check/drop 'forall
      (tc:goto-parser-state tc:parse-forall-exp-formals))))

(define tc:parse-forall-exp-formals
  (lambda (token)
    ((case (tc:scheme-token->class token)
       ((lparen) tc:seen-forall-exp-lparen)
       ((identifier) tc:seen-forall-exp-unrestricted-formals)
       (else (tc:unexpected-error '(lparen identifier))))
     token)))

(define tc:seen-forall-exp-lparen
  (lambda (token)
    (tc:check/drop 'lparen
      (tc:process-nt tc:parse-formals-guts
	(tc:check/drop 'rparen
	  (tc:process-nt tc:parse-scheme-exp
	    (tc:check/drop 'rparen
	      (tc:reduce 'forall-exp))))))))

(define tc:seen-forall-exp-unrestricted-formals
  (lambda (token)
    (tc:process-nt tc:parse-unrestricted-formals
      (tc:process-nt tc:parse-scheme-exp
	(tc:check/drop 'rparen
	  (tc:reduce 'forall-exp))))))

;;; sequences
(define tc:parse-sequence
  (lambda (token)
    (tc:process-nt tc:parse-nonempty-exp-list
      (tc:reduce 'sequence))))

;;; cond clauses
(define tc:parse-cond-clause
  (lambda (token)
    (tc:check/drop 'lparen
      (tc:goto-parser-state tc:parse-cond-clause-guts))))

(define tc:parse-cond-clause-guts
  (lambda (token)
    ((case (tc:scheme-token->class token)
       ((identifier)
	(case (tc:scheme-token->data token)
	  ((else) tc:cond-clause-seen-else)
	  (else tc:cond-clause-non-else)))
       (else tc:cond-clause-non-else))
     token)))

(define tc:cond-clause-non-else
  (lambda (token)
    (tc:process-nt tc:parse-scheme-exp
      (tc:goto-parser-state tc:cond-clause-decision))))

(define tc:cond-clause-decision
  (lambda (token)
    ((case (tc:scheme-token->class token)
       ((rparen) tc:cond-clause-parse-test-only)
       ((identifier)
	(case (tc:scheme-token->data token)
	  ((=>) tc:cond-clause-parse-recipient)
	  (else tc:cond-clause-parse-sequence)))
       (else tc:cond-clause-parse-sequence))
     token)))

(define tc:cond-clause-parse-sequence
  (lambda (token)
    (tc:process-nt tc:parse-sequence
      (tc:check/drop 'rparen
        (tc:reduce 'normal-cond-clause)))))

(define tc:cond-clause-parse-test-only
  (lambda (token)
    (tc:check/drop 'rparen
      (tc:reduce 'test-only-cond-clause))))
	
(define tc:cond-clause-parse-recipient
  (lambda (token)
    (tc:check/drop '=>
      (tc:process-nt tc:parse-scheme-exp
        (tc:check/drop 'rparen
          (tc:reduce 'test=>cond-clause))))))

(define tc:cond-clause-seen-else
  (lambda (token)
    (tc:check/drop 'else
      (tc:process-nt tc:parse-sequence
        (tc:check/drop 'rparen
	  (tc:reduce 'else-cond-clause))))))

;;; case clauses
(define tc:parse-case-clause
  (lambda (token)
    (tc:check/drop 'lparen
      (tc:goto-parser-state tc:parse-case-clause-guts))))

(define tc:parse-case-clause-guts
  (lambda (token)
    ((case (tc:scheme-token->class token)
       ((identifier)
	(case (tc:scheme-token->data token)
	  ((else) tc:case-clause-seen-else)
	  (else (tc:unexpected-error '(else lparen)))))
       ((lparen) tc:case-clause-non-else)
       (else (tc:unexpected-error '(lparen else))))
     token)))

(define tc:case-clause-non-else
  (lambda (token)
    (tc:check/drop 'lparen
      (tc:process-nt tc:parse-scheme-datum-list
        (tc:check/drop 'rparen
	  (tc:process-nt tc:parse-sequence
            (tc:check/drop 'rparen
              (tc:reduce 'normal-case-clause))))))))

(define tc:case-clause-seen-else
  (lambda (token)
    (tc:check/drop 'else
      (tc:process-nt tc:parse-sequence
        (tc:check/drop 'rparen
	  (tc:reduce 'else-case-clause))))))

;;; scheme-iteration-spec (for do)
(define tc:parse-scheme-iteration-spec
  (lambda (token)
    (tc:check/drop 'lparen
      (tc:check/shift 'identifier
	(tc:process-nt tc:parse-scheme-exp
	  (tc:goto-parser-state tc:parse-optional-iteration-step))))))

(define tc:parse-optional-iteration-step
  (lambda (token)
    ((case (tc:scheme-token->class token)
       ((rparen) tc:no-iteration-step)
       (else tc:iteration-step))
     token)))

(define tc:no-iteration-step
  (lambda (token)
    (tc:check/drop 'rparen
      (tc:reduce 'init-only-iteration-spec))))

(define tc:iteration-step
  (lambda (token)
    (tc:process-nt tc:parse-scheme-exp
      (tc:check/drop 'rparen
        (tc:reduce 'stepped-iteration-spec)))))

;;; cases clauses
(define tc:parse-cases-clause
  (lambda (token)
    (tc:check/drop 'lparen
      (tc:goto-parser-state tc:parse-cases-clause-guts))))

(define tc:parse-cases-clause-guts
  (lambda (token)
    ((case (tc:scheme-token->class token)
       ((identifier)
	(case (tc:scheme-token->data token)
	  ((else) tc:cases-clause-seen-else)
	  (else tc:cases-clause-non-else)))
       (else (tc:unexpected-error '(identifier else))))
     token)))

(define tc:cases-clause-non-else
  (lambda (token)
    (tc:check/shift 'identifier
      (tc:check/drop 'lparen
        (tc:process-nt tc:parse-no-dot-identifier-list
          (tc:check/drop 'rparen
	    (tc:process-nt tc:parse-sequence
              (tc:check/drop 'rparen
                (tc:reduce 'normal-cases-clause)))))))))

(define tc:parse-no-dot-identifier-list
  (lambda (token)
    ((tc:parse-loop* '(identifier) '(rparen)
       (lambda (action) (tc:check/shift 'identifier action)))
     token)))

(define tc:cases-clause-seen-else
  (lambda (token)
    (tc:check/drop 'else
      (tc:process-nt tc:parse-sequence
        (tc:check/drop 'rparen
	  (tc:reduce 'else-cases-clause))))))

;;; scheme-datum
(define tc:parse-scheme-datum
  (lambda (token)
    ((case (tc:scheme-token->class token)
      ((identifier) tc:seen-identifier-datum)
      ((number) tc:seen-number-datum)
      ((boolean) tc:seen-boolean-datum)
      ((string) tc:seen-string-datum)
      ((character) tc:seen-character-datum)
      ((lparen) tc:seen-lparen-datum)
      ((quote-char) tc:seen-quote-char-datum)
      ((quasiquote-char) tc:seen-quasiquote-char-datum)
      ((unquote-char) tc:seen-unquote-char-datum)
      ((unquote-splicing-char) tc:seen-unquote-splicing-char-datum)
      ((sharp-lparen) tc:seen-sharp-lparen-datum)
      (else (tc:unexpected-error
	     *tc:token-classes-that-start-scheme-datum*)))
     token)))

(define tc:seen-identifier-datum
  (lambda (token)
    (tc:check/shift 'identifier
      (tc:reduce 'simple-datum))))

(define tc:seen-number-datum
  (lambda (token)
    (tc:check/shift 'number
      (tc:reduce 'simple-datum))))

(define tc:seen-boolean-datum
  (lambda (token)
    (tc:check/shift 'boolean
      (tc:reduce 'simple-datum))))

(define tc:seen-string-datum
  (lambda (token)
    (tc:check/shift 'string
      (tc:reduce 'simple-datum))))

(define tc:seen-character-datum
  (lambda (token)
    (tc:check/shift 'character
      (tc:reduce 'simple-datum))))

(define tc:seen-lparen-datum
  (lambda (token)
    (tc:check/drop 'lparen
      (tc:goto-parser-state tc:parse-list-datum-guts))))

(define tc:parse-list-datum-guts
  (lambda (token)
    (tc:process-nt tc:parse-scheme-datum-list
      (tc:goto-parser-state tc:parse-optional-dotted-datum-list))))

(define tc:parse-scheme-datum-list
  (lambda (token)
    ((tc:parse-loop* *tc:token-classes-that-start-scheme-datum* '(dot rparen)
       (lambda (action) (tc:process-nt tc:parse-scheme-datum action)))
     token)))

(define tc:parse-optional-dotted-datum-list
  (lambda (token)
    ((case (tc:scheme-token->class token)
      ((rparen) tc:parse-end-of-normal-list)
      ((dot) tc:parse-optional-dotted-item)
      (else (tc:unexpected-error '(rparen dot))))
     token)))

(define tc:parse-end-of-normal-list
  (lambda (token)
    (tc:check/drop 'rparen
      (tc:reduce 'normal-list-datum))))

(define tc:parse-optional-dotted-item
  (lambda (token)
    (tc:check/drop 'dot
      (tc:process-nt tc:parse-scheme-datum
	(tc:check/drop 'rparen
	  (tc:reduce 'improper-list-datum))))))

(define tc:seen-sharp-lparen-datum
  (lambda (token)
    (tc:check/drop 'sharp-lparen
      (tc:process-nt tc:parse-scheme-vector-elements
	(tc:check/drop 'rparen
	  (tc:reduce 'vector-datum))))))

(define tc:parse-scheme-vector-elements
  (lambda (token)
    ((tc:parse-loop* *tc:token-classes-that-start-scheme-datum* '(rparen)
       (lambda (action) (tc:process-nt tc:parse-scheme-datum action)))
     token)))

(define tc:seen-quote-char-datum
  (lambda (token)
    (tc:check/drop 'quote-char
      (tc:process-nt tc:parse-scheme-datum
        (tc:reduce 'quoted-datum)))))

(define tc:seen-quasiquote-char-datum
  (lambda (token)
    (tc:check/drop 'quasiquote-char
      (tc:process-nt tc:parse-scheme-datum
        (tc:reduce 'quasiquoted-datum)))))

(define tc:seen-unquote-char-datum
  (lambda (token)
    (tc:check/drop 'unquote-char
      (tc:process-nt tc:parse-scheme-datum
        (tc:reduce 'unquoted-datum)))))

(define tc:seen-unquote-splicing-char-datum
  (lambda (token)
    (tc:check/drop 'unquote-splicing-char
      (tc:process-nt tc:parse-scheme-datum
        (tc:reduce 'unquote-splicing-datum)))))

;;; special cases for quasiquotations
(define tc:parse-scheme-quasiquoted-datum
  (lambda (token)
    ((case (tc:scheme-token->class token)
      ((identifier) tc:seen-identifier-datum)
      ((number) tc:seen-number-datum)
      ((boolean) tc:seen-boolean-datum)
      ((string) tc:seen-string-datum)
      ((character) tc:seen-character-datum)
      ((lparen) tc:seen-lparen-quasiquoted-datum)
      ((quote-char) tc:seen-quote-char-quasiquoted-datum)
      ((quasiquote-char) tc:seen-quasiquote-char-quasiquoted-datum)
      ((unquote-char) tc:seen-unquote-char-quasiquoted-datum)
      ((unquote-splicing-char) tc:seen-unquote-splicing-char-quasiquoted-datum)
      ((sharp-lparen) tc:seen-sharp-lparen-quasiquoted-datum)
      (else (tc:unexpected-error
	     *tc:token-classes-that-start-scheme-datum*)))
     token)))

(define tc:seen-lparen-quasiquoted-datum
  (lambda (token)
    (tc:check/drop 'lparen
      (tc:goto-parser-state tc:seen-lparen-quasiquoted-decision))))

(define tc:seen-lparen-quasiquoted-decision
  (lambda (token)
    ((case (tc:scheme-token->class token)
      ((identifier)
       (case (tc:scheme-token->data token)
	 ((unquote) tc:seen-unquote-quasiquoted-datum)
	 ((unquote-splicing) tc:seen-unquote-splicing-quasiquoted-datum)
	 ((quasiquote) tc:seen-quasiquote-quasiquoted-datum)
	 (else tc:parse-list-quasiquoted-datum-guts)))
      (else tc:parse-list-quasiquoted-datum-guts))
     token)))

(define tc:parse-list-quasiquoted-datum-guts
  (lambda (token)
    (tc:process-nt tc:parse-quasiquoted-datum-list
      (tc:goto-parser-state tc:parse-optional-dotted-quasiquoted-datum-list))))

(define tc:parse-quasiquoted-datum-list
  (lambda (token)
    ((tc:parse-loop* *tc:token-classes-that-start-scheme-datum* '(dot rparen)
       (lambda (action)
         (tc:process-nt tc:parse-scheme-quasiquoted-datum action)))
     token)))

(define tc:parse-optional-dotted-quasiquoted-datum-list
  (lambda (token)
    ((case (tc:scheme-token->class token)
      ((rparen) tc:parse-end-of-normal-list)
      ((dot) tc:parse-optional-dotted-quasiquoted-item)
      (else (tc:unexpected-error '(rparen dot))))
     token)))

(define tc:parse-optional-dotted-quasiquoted-item
  (lambda (token)
    (tc:check/drop 'dot
      (tc:process-nt tc:parse-scheme-quasiquoted-datum
	(tc:check/drop 'rparen
	  (tc:reduce 'improper-list-datum))))))

(define tc:seen-sharp-lparen-quasiquoted-datum
  (lambda (token)
    (tc:check/drop 'sharp-lparen
      (tc:process-nt tc:parse-scheme-quasiquoted-vector-elements
	(tc:check/drop 'rparen
	  (tc:reduce 'vector-datum))))))

(define tc:parse-scheme-quasiquoted-vector-elements
  (lambda (token)
    ((tc:parse-loop* *tc:token-classes-that-start-scheme-datum* '(rparen)
      (lambda (action)
        (tc:process-nt tc:parse-scheme-quasiquoted-datum action)))
     token)))

(define tc:seen-quote-char-quasiquoted-datum
  (lambda (token)
    (tc:check/drop 'quote-char
      (tc:process-nt tc:parse-scheme-quasiquoted-datum
        (tc:reduce 'quoted-datum)))))

(define tc:seen-quasiquote-char-quasiquoted-datum
  (lambda (token)
    (tc:check/drop 'quasiquote-char
      (tc:inc-qq-depth-and-process-nt! tc:parse-scheme-quasiquoted-datum
        (tc:dec-qq-depth-and-reduce! 'quasiquoted-datum)))))

(define tc:seen-quasiquote-quasiquoted-datum
  (lambda (token)
    (tc:check/drop 'quasiquote
      (tc:inc-qq-depth-and-process-nt! tc:parse-scheme-quasiquoted-datum
	(tc:check/drop 'rparen
          (tc:dec-qq-depth-and-reduce! 'quasiquoted-datum))))))

(define tc:seen-unquote-char-quasiquoted-datum
  (lambda (token)
    (tc:check/drop 'unquote-char
      (tc:dec-qq-depth-and-process-nt! tc:parse-scheme-exp
				       tc:parse-scheme-quasiquoted-datum
        (tc:inc-qq-depth-and-reduce! 'unquoted-exp
				     'unquoted-datum)))))

(define tc:seen-unquote-quasiquoted-datum
  (lambda (token)
    (tc:check/drop 'unquote
      (tc:dec-qq-depth-and-process-nt! tc:parse-scheme-exp
				       tc:parse-scheme-quasiquoted-datum
	(tc:check/drop 'rparen
          (tc:inc-qq-depth-and-reduce! 'unquoted-exp
				       'unquoted-datum))))))

(define tc:seen-unquote-splicing-char-quasiquoted-datum
  (lambda (token)
    (tc:check/drop 'unquote-splicing-char
      (tc:dec-qq-depth-and-process-nt! tc:parse-scheme-exp
				      tc:parse-scheme-quasiquoted-datum
        (tc:inc-qq-depth-and-reduce! 'unquote-splicing-exp
				     'unquote-splicing-datum)))))

(define tc:seen-unquote-splicing-quasiquoted-datum
  (lambda (token)
    (tc:check/drop 'unquote-splicing
      (tc:dec-qq-depth-and-process-nt! tc:parse-scheme-exp
				      tc:parse-scheme-quasiquoted-datum
	(tc:check/drop 'rparen
          (tc:inc-qq-depth-and-reduce! 'unquote-splicing-exp
		  		       'unquote-splicing-datum))))))

;;; input type expressions
(define tc:parse-input-type-exp
  (lambda (token)
    ((case (tc:scheme-token->class token)
       ((identifier) tc:seen-type-basic)
       ((lparen) tc:seen-type-lparen)
       (else (tc:unexpected-error '(identifier lparen))))
     token)))

(define tc:seen-type-basic
  (lambda (token)
    (tc:check/shift 'identifier
      (tc:reduce 'basic-input-type-exp))))

(define tc:seen-type-lparen
  (lambda (token)
    (tc:check/drop 'lparen
      (tc:goto-parser-state tc:parse-type-exp-guts))))

(define tc:parse-type-exp-guts
  (lambda (token)
    ((case (tc:scheme-token->class token)
       ((identifier)
	(let ((name (tc:scheme-token->data token)))
	  (case name
	    ((->) tc:seen-type-arrow)
	    ((forall) tc:seen-polymorphic-type)
	    ((all-of) tc:seen-intersection-type)
	    ((type-predicate-for) tc:seen-type-predicate-for-type)
	    ((variant-record) tc:seen-variant-record-type)
	    (else tc:seen-type-applied))))
       (else (tc:unexpected-error '(-> identifier))))
     token)))

(define tc:seen-type-arrow
  (lambda (token)
    (tc:check/drop '->
      (tc:check/drop 'lparen
	(tc:process-nt tc:input-type-exp-list
	  (tc:check/drop 'rparen
	    (tc:process-nt tc:parse-input-type-exp
              (tc:check/drop 'rparen
	        (tc:reduce 'function-input-type-exp)))))))))

(define tc:input-type-exp-list
  (lambda (token)
    ((tc:parse-loop* '(identifier lparen) '(rparen)
       (lambda (action) (tc:process-nt tc:parse-input-type-exp action)))
     token)))

(define tc:seen-intersection-type
  (lambda (token)
    (tc:check/drop 'all-of
      (tc:process-nt tc:input-type-exp-list
	(tc:check/drop 'rparen
	  (tc:reduce 'intersection-input-type-exp))))))

(define tc:seen-type-predicate-for-type
  (lambda (token)
    (tc:check/drop 'type-predicate-for
      (tc:process-nt tc:parse-input-type-exp
	(tc:check/drop 'rparen
	  (tc:reduce 'type-predicate-for-input-type-exp))))))

(define tc:seen-variant-record-type
  (lambda (token)
    (tc:check/drop 'variant-record
      (tc:process-nt tc:parse-variant-type-decl-list
	(tc:check/drop 'rparen
	  (tc:reduce 'variant-record-input-type-exp))))))

(define tc:seen-type-applied
  (lambda (token)
    (tc:check/shift 'identifier
      (tc:process-nt tc:input-type-exp-nonempty-list
	(tc:check/drop 'rparen
	  (tc:reduce 'applied-input-type-exp))))))

(define tc:input-type-exp-nonempty-list
  (lambda (token)
    ((tc:parse-loop+ '(identifier lparen) '(rparen)
      (lambda (action) (tc:process-nt tc:parse-input-type-exp action)))
     token)))

(define tc:seen-polymorphic-type
  (lambda (token)
    (tc:check/drop 'forall
      (tc:check/drop 'lparen
        (tc:process-nt tc:parse-no-dot-identifier-list
          (tc:check/drop 'rparen
	    (tc:process-nt tc:parse-input-type-exp
	      (tc:check/drop 'rparen
	        (tc:reduce 'polymorphic-input-type-exp)))))))))

;;; variant-type-decls
(define tc:parse-variant-type-decl-list
  (lambda (token)
    ((tc:parse-loop+ '(lparen) '(rparen)
      (lambda (action) (tc:process-nt tc:parse-variant-type-decl action)))
     token)))

(define tc:parse-variant-type-decl
  (lambda (token)
    (tc:check/drop 'lparen
      (tc:check/shift 'identifier
	(tc:process-nt tc:parse-field-type-bindings
	  (tc:check/drop 'rparen
	    (tc:reduce 'variant-type-decl)))))))

;;; field-type bindings
(define tc:parse-field-type-bindings
  (lambda (token)
    ((tc:parse-loop* '(lparen) '(rparen)
      (lambda (action) (tc:process-nt tc:parse-field-type-binding action)))
     token)))

(define tc:parse-field-type-binding
  (lambda (token)
    (tc:check/drop 'lparen
      (tc:check/shift 'identifier
	(tc:process-nt tc:parse-input-type-exp
	  (tc:check/drop 'rparen
	    (tc:reduce 'field-type-binding)))))))

;;; Parsed file cache ;;;;;

(define *tc:parsed-file-cache* '())

(deftype tc:reset-parsed-file-cache! (-> () void))
(define tc:reset-parsed-file-cache!
  (lambda ()
    (set! *tc:parsed-file-cache* '())))

(define-datatype tc:parsed-file-cache-entry tc:parsed-file-cache-entry?
  (tc:begin-parsing 
   (fname string?))
  (tc:parsed
   (fname string?) (program tc:scheme-program?)))

(define tc:parsed-file-cache-entry->fname
  (lambda (pfce)
    (cases tc:parsed-file-cache-entry pfce
      (tc:begin-parsing (fname) fname)
      (tc:parsed (fname program) fname))))

(deftype tc:begin-parsing-file! (-> (string) void))
(define tc:begin-parsing-file!
  (lambda (fname)
    (set! *tc:parsed-file-cache*
	  (cons (tc:begin-parsing fname) *tc:parsed-file-cache*))))

(deftype tc:done-parsing-file! (-> (string tc:scheme-program) void))
(define tc:done-parsing-file!
  (lambda (fname program)
    (set! *tc:parsed-file-cache*
	  (map (lambda (ce)
		 (if (string=? fname
			       (tc:parsed-file-cache-entry->fname ce))
		     (tc:parsed fname program)
		     ce))
	       *tc:parsed-file-cache*))))

(deftype tc:look-in-parsed-file-cache-for
  (-> (string tc:position) (maybe-of tc:scheme-program)))
(define tc:look-in-parsed-file-cache-for
  (lambda (filename position)
    (letrec ((look
	      (lambda (entries)
		(if (null? entries)
		    (make-nothing)
		    (cases tc:parsed-file-cache-entry (car entries)
		      (tc:begin-parsing (fname)
			(if (string=? filename fname)
			    (error (string-append
				    (tc:position->error-string position)
				    "infinitely recursive load of")
				   fname)
			    (look (cdr entries))))
		      (tc:parsed (fname program)
			(if (string=? filename fname)
			    (make-something program)
			    (look (cdr entries))))
		      (else (error "tc:parsed-file-cache-parsed?, missing case"
				   (car entries))))))))
      (look *tc:parsed-file-cache*))))
	
(deftype tc:remove-parsing-record-for! (-> (string) void))
(define tc:remove-parsing-record-for!
  (lambda (filename)
    (letrec ((rem
	      (lambda (entries)
		(if (null? entries)
		    '()
		    (cases tc:parsed-file-cache-entry (car entries)
		      (tc:begin-parsing (fname)
			(if (string=? filename fname)
			    (cdr entries)
			    (cons (car entries) (rem (cdr entries)))))
		      (else (cons (car entries) (rem (cdr entries)))))))))
      (set! *tc:parsed-file-cache* (rem *tc:parsed-file-cache*)))))

) ;; end module
